package com.jalios.jcms;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.io.Writer;
import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.sql.Connection;
import java.sql.DriverManager;
import java.text.DateFormat;
import java.text.DecimalFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Collection;
import java.util.Comparator;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.SortedSet;
import java.util.Stack;
import java.util.StringTokenizer;
import java.util.TreeSet;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

import javax.servlet.ServletContext;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.collections.ComparatorUtils;
import org.apache.commons.io.filefilter.TrueFileFilter;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.NDC;
import org.apache.oro.text.perl.Perl5Util;
import org.apache.oro.text.regex.MalformedPatternException;
import org.apache.oro.text.regex.MatchResult;
import org.apache.oro.text.regex.Pattern;
import org.apache.oro.text.regex.PatternCompiler;
import org.apache.oro.text.regex.PatternMatcher;
import org.apache.oro.text.regex.PatternMatcherInput;
import org.apache.oro.text.regex.Perl5Compiler;
import org.apache.oro.text.regex.Perl5Matcher;
import org.apache.oro.text.regex.Perl5Substitution;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.Namespace;

import com.jalios.io.IOUtil;
import com.jalios.io.RegexpFileFilter;
import com.jalios.io.ReverseFileFilter;
import com.jalios.jcms.context.JcmsContext;
import com.jalios.jcms.db.DBConstants;
import com.jalios.jcms.db.HibernateUtil;
import com.jalios.jcms.i18n.I18nUtil;
import com.jalios.jcms.logging.JcmsPatternLayout;
import com.jalios.jcms.portlet.PortalElement;
import com.jalios.jcms.workspace.Workspace;
import com.jalios.jspengine.FakedHttpSession;
import com.jalios.jspengine.FakedServletContext;
import com.jalios.jspengine.FakedServletRequest;
import com.jalios.jspengine.FakedServletResponse;
import com.jalios.jspengine.TemplateEngine;
import com.jalios.jstore.BasicStorable;
import com.jalios.jstore.StoreUtil;
import com.jalios.util.DateUtil;
import com.jalios.util.HtmlUtil;
import com.jalios.util.JProperties;
import com.jalios.util.LangProperty;
import com.jalios.util.ObjectLongTreeMap;
import com.jalios.util.ServletUtil;
import com.jalios.util.Util;
import com.jalios.util.XmlUtil;

/**
 * This class contains a set of utility static methods.
 *
 * @author Olivier Dedieu <olivier.dedieu@jalios.com>
 * @version $Revision: 97249 $
 */
public class JcmsUtil implements JcmsConstants {
  public static final String REVISION = "$Revision: 97249 $";

  private static final Logger logger = Logger.getLogger(JcmsUtil.class);

  public static final String DELIMITERS = ",; \t\n";

  static boolean forbiddenNoPattern = false;
  static Pattern forbiddenPattern;

  static Properties revisionCache = new Properties();
  static Properties releasedRevisionCache;
  static ArrayList<Class<?>> classList;
  // @formatter:off
  static String[] jarNames = new String[] {
    "jcms.jar", 
    "jaliosutil.jar", 
    "jstore.jar",
    "jtaglib.jar",
    "jdring.jar"
  };
  // @formatter:on

  private static String ignoreFileRegexp;
  private static String ignoreSignRegexp;

  private static Object jspEngineLock = new Object();

  private static final Map<Class<?>, String> ICON_CACHE = new HashMap<Class<?>, String>();

  private static final Channel channel = Channel.getChannel();

  // ---------------------------------------------------------------------------------------------------------------
  // Servlet Utility Method |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
  // ---------------------------------------------------------------------------------------------------------------

  /**
   * Returns true if the file is authorized for JSP inclusion.<p>
   * It must match the following criteria :
   * <ul>
   * <li>be a valid and existing file</li>
   * <li>be a JSP (extension == "jsp")</li>
   * <li>belongs to webapp directory</li>
   * <li>do not belongs to WEB-INF directory</li>
   * <li>do not belongs to upload directory</li>
   * </ul>
   * <p>
   * If property <code>channel.security.authorized-jsp-whitelist</code> is set to true,
   * the previous blacklist criteria are still verified, but relative path is only
   * authorized if its value is one of the value specified value through
   * properties <code>channel.security.authorized-jsp.*:</code>
   * 
   * @param relativePath the path to be tested
   * @return true if the given file is a file suitable for dynamic inclusion in a JSP
   */
  public static boolean isLegalFile(String relativePath) {
    if (Util.isEmpty(relativePath)) {
      return false;
    }

    File file = new File(channel.getRealPath(relativePath));

    // Exists and is a valid file
    if (!file.exists() || !file.isFile()) {
      return false;
    }

    // Is a .jsp, and nothing else
    if (!"jsp".equalsIgnoreCase(IOUtil.getExtension(file))) {
      return false;
    }

    // belong to webapp root directory
    File webappDir = new File(channel.getWebappPath());
    if (!IOUtil.belongsToDirectory(webappDir, file)) {
      return false;
    }

    // is NOT under WEB-INF directory
    File webinfDir = new File(channel.getWebInfPath("/"));
    if (IOUtil.belongsToDirectory(webinfDir, file)) {
      return false;
    }

    // is NOT under upload directory
    File uploadDir = new File(channel.getUploadPath("/"));
    if (IOUtil.belongsToDirectory(uploadDir, file)) {
      return false;
    }

    // Fix bug JCMS-3781/SEC-3 : JSP blacklist can be bypassed using path traversal parameter
    // --> Make sure that relativePath is sanitize and any path traversal is removed
    relativePath = IOUtil.getRelativePath(webappDir, file);

    // Fix bug JCMS-3766/SEC-1 : StackOverflowError followed by java.lang.OutOfMemoryError: Java heap space
    // --> Always forbid display.jsp
    if (relativePath.matches("display.jsp")) {
      return false;
    }

    // Improvement JCMS-3290 : Allow properties to blacklist jsp call with parameter "jsp="
    // Explicit Blacklist version : if enabled, JSP declared in those properties will be forbidden
    boolean useExplicitBlacklist = Util.toBoolean(channel.getProperty("channel.security.forbidden-jsp-blacklist", null), false);
    if (useExplicitBlacklist) {
      JProperties jspRegexProps = channel.getProperties("channel.security.forbidden-jsp.");
      if (jspRegexProps != null) {
        for (String forbiddenJspPath : jspRegexProps.values()) {
          if (relativePath.equals(forbiddenJspPath) || relativePath.matches(forbiddenJspPath)) {
            if (logger.isDebugEnabled()) {
              logSecurityIssue(logger, Level.DEBUG,
                  "'" + relativePath + "' is not a valid JSP file authorized for dynamic inclusion, it has been blacklisted by property");
            }
            return false;
          }
        }
      }
    }

    // Whitelist version : if enabled, only JSP declared in those properties will be allowed
    boolean useWhitelist = Util.toBoolean(channel.getProperty("channel.security.authorized-jsp-whitelist", null), false);
    if (useWhitelist) {
      JProperties jspRegexProps = channel.getProperties("channel.security.authorized-jsp.");
      if (jspRegexProps != null) {
        for (String authorizedJspPath : jspRegexProps.values()) {
          if (relativePath.equals(authorizedJspPath)) {
            return true;
          }
        }
      }
      if (logger.isDebugEnabled()) {
        logSecurityIssue(logger, Level.DEBUG, "'" + relativePath + "' is not a valid JSP file authorized for dynamic inclusion.");
      }
      return false;
    }

    return true;
  }

  // ---------------------------------------------------------------------------------------------------------------
  // JCMS Utility Method ||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
  // ---------------------------------------------------------------------------------------------------------------

  /**
   * Returns a CSV view of the given boolean value
   *
   * @param lang the language of the output
   * @param bool the boolean value
   * @return a CSV view of the given boolean value
   * @since jcms-5.7.0
   */
  public static String convertToCSV(String lang, boolean bool) {
    return bool ? "o" : "-";
  }

  /**
   * Returns a CSV view of the given double value
   *
   * @param lang the language of the output
   * @param value the double value
   * @return a CSV view of the given boolean value
   * @since jcms-9.0.5
   */
  public static String convertToCSV(String lang, long value) {
    return String.valueOf(value);
  }

  /**
   * Returns a CSV view of the given double value
   *
   * @param lang the language of the output
   * @param value the double value
   * @return a CSV view of the given boolean value
   * @since jcms-9.0.5
   */
  public static String convertToCSV(String lang, double value) {
    return new DecimalFormat("#0.00").format(value);
  }

  /**
   * Returns a CSV view of the given object in the given language.
   *
   * @param lang the language of the output
   * @param obj the object
   * @return a CSV view of the given object
   * @since jcms-5.7.0
   */
  public static String convertToCSV(String lang, Object obj) {
    return "\"" + convertToCSVRecursive(lang, obj) + "\"";
  }

  /**
   * Returns a CSV view of the given object in the specified language.
   *
   * @param lang the language of the output
   * @param obj the object
   * @param displayObjectId if true, the object's id will be appended to the data's name.
   * @return a CSV view of the given object
   */
  public static String convertToCSV(String lang, Object obj, boolean displayObjectId) {
    return "\"" + convertToCSVRecursive(lang, obj, false, displayObjectId) + "\"";
  }

  static String convertToCSVRecursive(String lang, Object obj) {
    return convertToCSVRecursive(lang, obj, true, false);
  }

  static String convertToCSVRecursive(String lang, Object obj, boolean displayDataNameAndId, boolean displayDataId) {
    if (obj == null) {
      return "";
    }
    if (lang == null) {
      lang = channel.getLanguage();
    }

    // String
    if (obj instanceof String) {
      String value = Util.getString(obj, "");
      value = value.replace("\"", "\"\""); // Fix bug JCMS-2036
      return value;
    }

    // Data
    // - don't forget to use getDataName and language to fix bug JCMS-2050
    // - also invoke convertToCSVRecursive again to make sure String are escaped properly to fix bug JCMS-2036
    if (obj instanceof Data) {
      Data data = (Data) obj;
      if (displayDataNameAndId) {
        return convertToCSVRecursive(lang, data.getDataName(lang)) + " [" + data.getId() + "]";
      } else if (displayDataId) {
        return data.getId();
      } else {
        return convertToCSVRecursive(lang, data.getDataName(lang));
      }
    }

    // Date
    if (obj instanceof Date) {
      return channel.getDateTimeFormat(lang).format((Date) obj);
    }

    // Collection
    if (obj instanceof Collection) {
      int cpt = 0;
      StringBuilder sb = new StringBuilder();
      for (Iterator<?> it = ((Collection<?>) obj).iterator(); it.hasNext(); cpt++) {
        sb.append(cpt > 0 ? "\n" : "").append(convertToCSVRecursive(lang, it.next(), displayDataNameAndId, displayDataId));
      }
      return sb.toString();
    }

    // Array
    if (obj.getClass().isArray()) {
      Object[] array = (Object[]) obj;
      StringBuilder sb = new StringBuilder();
      for (int i = 0; i < array.length; i++) {
        sb.append(i > 0 ? "\n" : "").append(convertToCSVRecursive(lang, array[i], displayDataNameAndId, displayDataId));
      }
      return sb.toString();
    }

    // Primitive
    if (obj.getClass().isPrimitive()) {
      return obj.toString();
    }

    // Class linked to primitive
    List<Class> classesList = Arrays
        .asList(new Class[] { Boolean.class, Byte.class, Character.class, Short.class, Integer.class, Long.class, Float.class, Double.class });
    if (classesList.contains(obj.getClass())) {
      return obj.toString();
    }

    return "";
  }

  /**
   * Export a dataSet to CSV
   *
   * @param dataSet the set of data to be exported
   * @param lang the language
   * @param writer the writer where to print the export
   * @since jcms-5.0.0
   */
  public static void exportCSV(Set<? extends Data> dataSet, String lang, Writer writer) {
    if (Util.isEmpty(dataSet) || writer == null) {
      return;
    }
    PrintWriter pw = new PrintWriter(writer);

    Class clazz = null;
    // Print header
    try {
      clazz = dataSet.iterator().next().getClass();
      java.lang.reflect.Method method = clazz.getMethod("getCSVHeader", new Class[] { String.class });
      clazz = method.getDeclaringClass();
      pw.println(method.invoke(null, new Object[] { lang }));
    } catch (Exception ex) {
      logger.warn("exportCSV: Cannot print header", ex);
    }

    // Print data
    for (Data data : dataSet) {
      // If clazz not null, remove object not of the class clazz

      if (clazz != null && !clazz.isAssignableFrom(data.getClass())) {
        logger.warn("method exportCSV has been called on an non homogeneous set " + dataSet);
      } else {
        pw.println(data.toCSV(lang));
      }
    }
  }

  /**
   * Export a Data Integrity Report to CSV
   * 
   * @param lang the language
   * @param writer the writer where to print the export
   * @since jcms-5.5.0
   */
  public static void exportCSVIntegrityReport(String lang, Writer writer) {
    Map<Data, ControllerStatus> report = channel.getDataIntegrityReport(false);
    exportCSVIntegrityReport(report, lang, writer);
  }

  /**
   * Export a Data Integrity Report to CSV
   * 
   * @param report the report to be exported
   * @param lang the language
   * @param writer the writer where to print the export
   * @since jcms-5.5.0
   */
  public static void exportCSVIntegrityReport(Map<Data, ControllerStatus> report, String lang, Writer writer) {
    if (writer == null) {
      return;
    }

    PrintWriter pw = new PrintWriter(writer);

    // Print header
    pw.print(convertToCSV(lang, glp(lang, "ui.com.lbl.id")));
    pw.print(";");
    pw.print(convertToCSV(lang, glp(lang, "ui.adm.dic.lbl.type")));
    pw.print(";");
    pw.print(convertToCSV(lang, glp(lang, "ui.adm.dic.lbl.value")));
    pw.print(";");
    pw.println(convertToCSV(lang, glp(lang, "ui.adm.dic.lbl.const")));

    // Print data
    for (Map.Entry<Data, ControllerStatus> entry : report.entrySet()) {
      Data itData = entry.getKey();
      ControllerStatus itStatus = entry.getValue();

      pw.print(convertToCSV(lang, itData.getId()));
      pw.print(";");
      pw.print(convertToCSV(lang, Util.getClassShortName(itData)));
      pw.print(";");
      pw.print(convertToCSV(lang, itData.getDataName(lang)));
      pw.print(";");
      pw.println(convertToCSV(lang, itStatus.getMessage(lang)));
    }
  }

  /**
   * Return a Set of classes of Types sorted alphabetically according to the title.
   * Two classes with the same label are sorted with the technical name of the java class.
   * Labels are localized according the given language.
   * 
   * @param subTypeList the collection that contain Class
   * @param workspace the workspace used to get the label associated to the class.
   * @param userLang the current language of the user logged
   * @return a Map Label/Class of Types sorted alphabetically.
   */
  public static TreeSet getSortedClass(Collection subTypeList, final Workspace workspace, final String userLang) {
    Comparator<Class<?>> comparator = new Comparator<Class<?>>() {
      public int compare(Class<?> c1, Class<?> c2) {
        String itClass1Label = Util.isEmpty(workspace) ? channel.getTypeLabel(c1, userLang) : workspace.getTypeLabel(c1, userLang);
        String itClass2Label = Util.isEmpty(workspace) ? channel.getTypeLabel(c2, userLang) : workspace.getTypeLabel(c2, userLang);
        if (Util.isEmpty(itClass1Label) || Util.isEmpty(itClass2Label) || itClass1Label.equals(itClass2Label)) {
          return c1.getName().compareTo(c2.getName());
        }
        String t1 = Util.unaccentuate(itClass1Label);
        String t2 = Util.unaccentuate(itClass2Label);
        return t1.compareToIgnoreCase(t2); // cannot be 0 as equality is tested above.
      }
    };

    TreeSet clazzSet = new TreeSet(comparator);
    if (Util.notEmpty(subTypeList)) {
      clazzSet.addAll(subTypeList);
    }
    return clazzSet;
  }

  /**
   * Quietly returns the Class matching the given className or defClass if it doesn't exists.
   * 
   * @param className the name of the class
   * @param defClass the default class if class name do not match
   * 
   * @return Class the matching class
   */
  public static Class getClass(String className, Class defClass) {
    if (Util.isEmpty(className)) {
      return defClass;
    }

    try {
      Class clazz = channel.getClass(className);
      if (clazz != null) {
        return clazz;
      }
    } catch (ClassNotFoundException ex) {
      logger.warn("An exception occured while retrieving class from class name \"" + className + "\"", ex);
      return defClass;
    }

    return defClass;
  }

  // Challenge : make the method getSortedClass() work for most of the cases below :
  /*
  static {
    Set<Class> ret0;
    Set<Class<?>> ret1;
    TreeSet<Class> ret2;
    TreeSet<Class<?>> ret3;
    TreeSet<Class<Data>> ret4;
    
    List list0 = new ArrayList();
    getSortedClass(list0, null, null); 
    ret0 = getSortedClass(list0, null, null); 
    ret1 = getSortedClass(list0, null, null); 
    ret2 = getSortedClass(list0, null, null); 
    ret3 = getSortedClass(list0, null, null); 
    ret4 = getSortedClass(list0, null, null); 
    
    List<Class> list1 = new ArrayList<Class>();
    getSortedClass(list1, null, null); 
    ret0 = getSortedClass(list1, null, null); 
    ret1 = getSortedClass(list1, null, null); 
    ret2 = getSortedClass(list1, null, null); 
    ret3 = getSortedClass(list1, null, null); 
    ret4 = getSortedClass(list1, null, null); 
    
    List<Class<?>> list2 = new ArrayList<Class<?>>();
    getSortedClass(list2, null, null); 
    ret0 = getSortedClass(list2, null, null); 
    ret1 = getSortedClass(list2, null, null); 
    ret2 = getSortedClass(list2, null, null); 
    ret3 = getSortedClass(list2, null, null); 
    ret4 = getSortedClass(list2, null, null); 
    
    List<Class<Data>> list3 = new ArrayList<Class<Data>>();
    getSortedClass(list3, null, null); 
    ret0 = getSortedClass(list3, null, null); 
    ret1 = getSortedClass(list3, null, null); 
    ret2 = getSortedClass(list3, null, null); 
    ret3 = getSortedClass(list3, null, null); 
    ret4 = getSortedClass(list3, null, null); 
  }
  */

  /**
   * Return all the members represented by an Array of Group ids
   * 
   * @param gids an array of Group ids.
   * @return Set of Member
   */
  public static Set<Member> getMemberSetFromGroups(String[] gids) {
    if (Util.isEmpty(gids)) {
      return null;
    }

    Set<Member> memberSet = new TreeSet<Member>();
    for (int i = 0; i < gids.length; i++) {
      Group grp = channel.getGroup(gids[i]);
      if (grp == null) {
        continue;
      }
      memberSet.addAll(grp.getMemberSet());
    }
    return memberSet;
  }

  /**
   * Returns an array of {@link Member}s represented by the concatenation of an array of {@link Group}s and an array of {@link Member}s.
   * It is possible to filter the returned members by defining members to delete in the last parameter.
   * 
   * @param groups array of {@link Group}s
   * @param members array of {@link Member}s
   * @param removeMembers {@link Member}s to remove in the final array of {@link Member}s
   * @return array of {@link Member}s
   */
  public static Member[] getAllMembers(Group[] groups, Member[] members, Member... removeMembers) {
    Set<Member> allMembers = new LinkedHashSet<Member>();
    if (Util.notEmpty(groups)) {
      for (Group group : groups) {
        if (group != null) {
          allMembers.addAll(group.getMemberSet());
        }
      }
    }
    if (Util.notEmpty(members)) {
      for (Member member : members) {
        if (member != null) {
          allMembers.add(member);
        }
      }
    }
    // Remove given members
    if (Util.notEmpty(removeMembers)) {
      if (allMembers.size() > removeMembers.length) {
        for (Member member : removeMembers) {
          remove(allMembers, member);
        }
      } else {
        for (Iterator<Member> it = allMembers.iterator(); it.hasNext();) {
          Member member = it.next();
          if (contains(removeMembers, member)) {
            it.remove();
          }
        }
      }
    }
    return allMembers.toArray(new Member[] {});
  }

  private static void remove(Set<Member> allMembers, Member member) {
    if (allMembers == null || member == null) {
      return;
    }
    for (Iterator<Member> it = allMembers.iterator(); it.hasNext();) {
      Member mbr = it.next();
      if (mbr.getId().equals(member.getId())) {
        it.remove();
        break;
      }
    }
  }

  private static boolean contains(Member[] members, Member member) {
    if (members == null || member == null) {
      return false;
    }
    for (Member mbr : members) {
      if (mbr != null && mbr.getId().equals(member.getId())) {
        return true;
      }
    }
    return false;
  }

  /**
   * Convert a String of Data ids into a List of Data.
   * 
   * @param str of id seperated with specified delimiter ie.
   * @param delim a delimiter (required, no default value is used)
   * @return List a list of Data, never return null.
   * @see #dataListToString(Collection, String) to perform the opposite operation
   * @see #stringToDataList(String, String, Class) for generic method
   * @since jcms-4.0
   */
  public static List<Data> stringToDataList(String str, String delim) {
    return stringToDataList(str, delim, null);
  }

  /**
   * Convert a String of Data ids into a List of Data.
   * 
   * @param <T> any type of Data to retrieve
   * @param str of id seperated with specified delimiter ie.
   * @param delim a delimiter (required, no default value is used)
   * @param clazz the Class of object to retrieve
   * @return List a list of Data, never return null.
   * @see #dataListToString(Collection, String) to perform the opposite operation
   * @since jcms-6.0
   */
  public static <T extends Data> List<T> stringToDataList(String str, String delim, Class<T> clazz) {
    List<T> list = new ArrayList<T>();

    if (Util.isEmpty(str)) {
      return list;
    }

    String[] split = Util.split(str, delim);
    for (int i = 0; i < split.length; i++) {
      String dataId = split[i] != null ? split[i].trim() : null;
      Data data = channel.getData(dataId);
      if (data != null && (clazz == null || clazz.isInstance(data))) {
        @SuppressWarnings("unchecked")
        T t = (T) data;
        list.add(t);
      }
    }

    return list;
  }

  /**
   * Convert a Collection of Data into a String composed by id, separated by the specified delimiter.
   * <p>
   * Example :
   * 
   * <pre>
   * List&lt;Data> list = new ArrayList&lt;Data>();
   * list.add(channel.getDefaultAdmin());
   * list.add(channel.getDefaultGroup());
   * String ids = dataListToString(list, ",");
   * </pre>
   * 
   * will return the following string : <code>"j2,j_36"</code>
   * 
   * @param <T> any class extending Data
   * @param list a Collection of Data
   * @param delim a delimiter (required, no default value is used)
   * @return String a string of id seperated with specified delimiter ie.
   * @see #stringToDataList(String, String) to perform the opposite operation
   * @since jcms-4.0
   */
  public static <T extends Data> String dataListToString(Collection<T> list, String delim) {
    if (list == null) {
      return "";
    }

    StringBuffer sb = new StringBuffer();
    boolean ok = false;
    for (Iterator<T> iter = list.iterator(); iter.hasNext(); ok = true) {
      if (ok) {
        sb.append(delim);
      }
      Data data = iter.next();
      if (data != null) {
        sb.append(data.getId());
      }
    }
    return sb.toString();
  }

  /**
   * Convert a Data[] to a String[] composed of
   * their name (label, title, name, ...)
   * 
   * @param datas Data[] an array of data to convert
   * @param lang the ISO-639 language code to be used to retrieve the localized Data's name
   * @return String[] an array of labels
   * @since jcms-5.7.0
   */
  public static String[] dataArrayToNameArray(Data[] datas, String lang) {
    if (datas == null) {
      return null;
    }

    String[] labels = new String[datas.length];
    for (int i = 0; i < datas.length; i++) {
      if (datas[i] != null) {
        labels[i] = datas[i].getDataName(lang);
      }
    }
    return labels;
  }

  /**
   * Convert a Data[] to a String[] composed by id
   * 
   * @param datas Data[] an array of data to convert
   * @return String[] an array of ids
   * @since jcms-5.0
   */
  public static String[] dataArrayToStringArray(Data[] datas) {
    if (datas == null) {
      return null;
    }

    String[] ids = new String[datas.length];
    for (int i = 0; i < datas.length; i++) {
      if (datas[i] != null) {
        ids[i] = datas[i].getId();
      }
    }
    return ids;
  }

  /**
   * Convert a Data[] to a String[] composed by id
   * 
   * @param ids an array of ids
   * @return String[]
   * @since jcms-5.0
   */
  public static Data[] stringArrayToDataArray(String[] ids) {
    if (ids == null) {
      return null;
    }
    Data[] datas = new Data[ids.length];
    for (int i = 0; i < ids.length; i++) {
      if (ids[i] != null) {
        datas[i] = channel.getData(ids[i]);
      }
    }
    return datas;
  }

  /**
   * Convert a Data[] to a String[] composed by id
   * 
   * @param <T> any Data type
   * @param clazz the Data class
   * @param ids an array of ids
   * @return String[]
   * @since jcms-5.0
   */
  public static <T extends Data> T[] stringArrayToDataArray(Class<T> clazz, String[] ids) {
    if (ids == null) {
      return null;
    }
    T[] datas = (T[]) Array.newInstance(clazz, ids.length);
    for (int i = 0; i < ids.length; i++) {
      if (ids[i] != null) {
        datas[i] = channel.getData(clazz, ids[i]);
      }
    }
    return datas;
  }

  /**
   * Returns the set of ids contained in the given collection of data.
   * 
   * @param <T> any Data type
   * @param dataColl a collection of Data
   * @return the set of ids contained in the given collection of data (never return null).
   * @since jcms-6.0.0
   * @see #idCollectionToDataSet(Collection, Class) to perform the opposite operation
   */
  public static <T extends Data> Set<String> dataCollectionToIdSet(Collection<T> dataColl) {
    Set<String> set = new LinkedHashSet<String>();

    if (dataColl == null) {
      return set;
    }

    for (Data data : dataColl) {
      if (data == null) {
        continue;
      }
      set.add(data.getId());
    }
    return set;
  }

  /**
   * Returns the list of ids contained in the given collection of data.
   * 
   * @param <T> any Data type
   * @param dataColl a collection of Data
   * @return the list of ids contained in the given collection of data (never return null).
   * @since jcms-6.1.0
   * @see #idCollectionToDataList(Collection, Class) to perform the opposite operation
   */
  public static <T extends Data> List<String> dataCollectionToIdList(Collection<T> dataColl) {
    List<String> list = new ArrayList<String>();

    if (dataColl == null) {
      return list;
    }

    for (Data data : dataColl) {
      if (data == null) {
        continue;
      }
      list.add(data.getId());
    }
    return list;
  }

  /**
   * Convert a Data[] to a Set&lt;String&gt; composed by id
   * 
   * @param datas Data[] an array of data to convert
   * @return the set of ids contained in the given array of data (never return null).
   * @since jcms-6.1.0
   * @see #idCollectionToDataSet(Collection, Class) to perform the opposite operation
   */
  public static Set<String> dataArrayToIdSet(Data[] datas) {
    Set<String> set = new HashSet<String>();
    if (datas == null) {
      return set;
    }

    for (int i = 0; i < datas.length; i++) {
      if (datas[i] != null) {
        String id = datas[i].getId();
        if (Util.notEmpty(id)) {
          set.add(id);
        }
      }
    }
    return set;
  }

  /**
   * Returns a Set of Data from a Collection of data's id.
   * 
   * @param <T> any type of Data to retrieve
   * @param idColl a Collection of data id
   * @param clazz the Class of object to retrieve
   * @return a new Set of Data (never return null).
   * @since jcms-6.1.0
   * @see #dataCollectionToIdSet(Collection) to perform the opposite operation
   */
  public static <T extends Data> Set<T> idCollectionToDataSet(Collection<String> idColl, Class<T> clazz) {
    return idCollectionToDataLinkedHashSet(idColl, clazz);
  }

  /**
   * Returns a LinkedHashSet of Data from a Collection of data's id.
   * 
   * @param <T> any type of Data to retrieve
   * @param idColl a Collection of data id
   * @param clazz the Class of object to retrieve
   * @return a new LinkedHashSet of Data (never return null).
   * @since jcms-6.2.0
   * @see #dataCollectionToIdSet(Collection) to perform the opposite operation
   */
  public static <T extends Data> LinkedHashSet<T> idCollectionToDataLinkedHashSet(Collection<String> idColl, Class<T> clazz) {
    LinkedHashSet<T> set = new LinkedHashSet<T>();

    if (Util.isEmpty(idColl)) {
      return set;
    }

    fillIdCollectionWithDataCollection(idColl, clazz, set);
    return set;
  }

  /**
   * Returns a TreeSet of Data from a Collection of data's id.
   * 
   * @param <T> any type of Data to retrieve
   * @param idColl a Collection of data id
   * @param clazz the Class of object to retrieve
   * @return a new TreeSet of Data (never return null).
   * @since jcms-6.2.0
   * @see #dataCollectionToIdSet(Collection) to perform the opposite operation
   */
  public static <T extends Data> TreeSet<T> idCollectionToDataTreeSet(Collection<String> idColl, Class<T> clazz) {
    TreeSet<T> set = new TreeSet<T>();

    if (Util.isEmpty(idColl)) {
      return set;
    }

    fillIdCollectionWithDataCollection(idColl, clazz, set);
    return set;
  }

  /**
   * Returns a HashSet of Data from a Collection of data's id.
   * 
   * @param <T> any type of Data to retrieve
   * @param idColl a Collection of data id
   * @param clazz the Class of object to retrieve
   * @return a new HashSet of Data (never return null).
   * @since jcms-6.2.0
   * @see #dataCollectionToIdSet(Collection) to perform the opposite operation
   */
  public static <T extends Data> HashSet<T> idCollectionToDataHashSet(Collection<String> idColl, Class<T> clazz) {
    HashSet<T> set = new HashSet<T>();

    if (Util.isEmpty(idColl)) {
      return set;
    }

    fillIdCollectionWithDataCollection(idColl, clazz, set);
    return set;
  }

  /**
   * Returns a List of Data from a Collection of data's id.
   * 
   * @param <T> any type of Data to retrieve
   * @param idColl a Collection of data id
   * @param clazz the Class of object to retrieve
   * @return a new List of Data (never return null).
   * @since jcms-6.1.0
   * @see #dataCollectionToIdList(Collection) to perform the opposite operation
   */
  public static <T extends Data> List<T> idCollectionToDataList(Collection<String> idColl, Class<T> clazz) {
    List<T> list = new ArrayList<T>();

    if (Util.isEmpty(idColl)) {
      return list;
    }

    fillIdCollectionWithDataCollection(idColl, clazz, list);
    return list;
  }

  /**
   * Fill a Data collection with Data retrieved from a collection of ids.
   * <p>
   * Invalid or null values are not inserted in the collection.
   * 
   * @param idColl a collection of Data id
   * @param clazz the class that Data must match in order to be added to the collection of Data
   * @param dataColl the collection in which to add the Data retrieved MUST not be null
   */
  public static <T extends Data> void fillIdCollectionWithDataCollection(Collection<String> idColl, Class<T> clazz, Collection<T> dataColl) {
    for (String id : idColl) {
      Data data = channel.getData(id);
      if (data != null && (clazz == null || clazz.isInstance(data))) {
        @SuppressWarnings("unchecked")
        T t = (T) data;
        dataColl.add(t);
      }
    }
  }

  public static <T extends Data> T[] idCollectionToDataArray(Collection<String> idColl, Class<T> clazz) {
    if (Util.isEmpty(idColl)) {
      return null;
    }

    T[] array = (T[]) Array.newInstance(clazz, idColl.size());
    int i = 0;
    for (String id : idColl) {
      array[i++] = channel.getData(clazz, id);
    }
    return array;

  }

  /**
   * Joins a collection of Object with the given separator into a
   * String. Uses the <code>getDataName(userLang)</code> method for each iterated
   * Data.
   *
   * @param coll the collection
   * @param sep the separator
   * @param userLang the ISO-639 language code to be used to retrieve localized titles
   * @return a String which contains all the iterated objects.
   * @see com.jalios.util.Util#join(Iterator,String,int,String)
   * @since jcms-4.0
   */
  public static String join(Collection<? extends Data> coll, String sep, String userLang) {
    if (Util.isEmpty(coll)) {
      return "";
    }

    return join(coll.iterator(), sep, userLang);
  }

  /**
   * Joins a collection of Object with the given separator into a
   * String. Uses the <code>getDataName(userLang)</code> method for each iterated
   * Data.
   *
   * @param it the iterator of the collection
   * @param sep the separator
   * @param userLang the ISO-639 language code to be used to retrieve localized titles
   * @return a String which contains all the iterated objects.
   * @see com.jalios.util.Util#join(Iterator,String,int,String)
   * @since jcms-4.0
   */
  public static String join(Iterator it, String sep, String userLang) {

    StringBuffer sb = new StringBuffer();
    for (; it.hasNext();) {
      Object obj = it.next();
      if (!(obj instanceof Data)) {
        continue;
      }
      Data data = (Data) obj;
      sb.append(data.getDataName(userLang) + sep);
    }

    String str = sb.toString();

    // Remove the trailing separator
    if (str.length() > 0) {
      str = str.substring(0, str.length() - sep.length());
    }
    return str;
  }

  /**
   * This method will expand each String[] keys
   * of a given map to a given length.
   * 
   * @param mlMap Map of key / String[]
   * @param main Map of key / String[] used to calculate String[] length
   * @return Map the modified mlMap
   */
  public static Map expandArray(Map mlMap, Map main) {

    int length = 0;

    for (Iterator it = main.entrySet().iterator(); it.hasNext();) {
      Map.Entry itEntry = (Map.Entry) it.next();
      String[] itValue = (String[]) itEntry.getValue();
      if (itValue == null)
        continue;
      length = itValue.length;
      break;
    }

    for (Iterator it = mlMap.entrySet().iterator(); it.hasNext();) {
      Map.Entry itEntry = (Map.Entry) it.next();
      String[] itValue = (String[]) itEntry.getValue();
      String[] newValue = new String[length];
      itEntry.setValue(newValue);

      if (itValue == null)
        continue;

      int cpylength = itValue.length > length ? length : itValue.length;
      System.arraycopy(itValue, 0, newValue, 0, cpylength);
    }

    return mlMap;
  }

  /**
   * This method will expand each String[] keys
   * of a given map to a given length.
   * 
   * @param array String[]
   * @param main String[] used to calculate String[] length
   * @return String[] a copy of the orifgianl array
   */
  public static String[] expandArray(String[] array, String[] main) {
    if (main == null)
      return null;
    int length = main.length;
    String[] newValue = new String[length];
    if (array == null)
      return newValue;
    int cpylength = main.length > length ? length : main.length;
    System.arraycopy(main, 0, newValue, 0, cpylength);
    return newValue;
  }

  // ---------------------------------------------------------------------------------------------------------------
  // Date Java Utility Method |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
  // ---------------------------------------------------------------------------------------------------------------

  /**
   * Return the next schedule date after/before reference date.
   * <br><br>
   * If Reference 04/05/1977 and schedule is 2 Month the method will forward to 04/07/1977
   * If Reference 04/05/1977 and schedule is 2 Month the method will backward to 04/03/1977
   * 
   * @param reference the reference starting date
   * @param forward forward to now or backward before now
   * @param schedule the time to add to reference date.
   * @return the next schedule Date, never return null
   */
  public static Date getScheduleDate(Date reference, long schedule, boolean forward) {
    return getScheduleDate(reference, reference, schedule, forward);
  }

  /**
   * Return the next schedule date after/before now (if reference date is before/after now)
   * <br><br>
   * If Reference 04/05/1977 and targetDate 04/08/1977 and schedule is 2 Month
   * the method will forward to 04/09/1977.
   * 
   * @param reference the reference starting date
   * @param targetDate the target date
   * @param schedule the time to add to reference date.
   * @param forward should forward or backward to targetDate
   * @return the next schedule Date, never return null
   */
  public static Date getScheduleDate(Date reference, Date targetDate, long schedule, boolean forward) {
    Calendar nowCal = Calendar.getInstance();
    nowCal.setTime(targetDate);

    Calendar refCal = Calendar.getInstance();
    refCal.setTime(reference);

    int[] modifiers = getScheduleModifier(schedule, forward);
    loopScheduleDate(refCal, nowCal, modifiers, forward, -1);

    return refCal.getTime();
  }

  /**
   * Calculate loops to do between reference date and target date then apply
   * loops on all toUpdate Dates.
   * <br>
   * long l1 is a schedule to apply
   * Date d1 is a schedule date
   * Date d2 is now
   * Date d3 and d4 are dates to loop on.
   * <br>
   * This method first calculate the quantity of schedule between d1 and d2
   * then apply this loop on d3 and d4.
   * 
   * @param reference the reference starting date
   * @param targetDate the target date
   * @param toUpdate the date to update
   * @param schedule the time to add to reference date.
   * @param forward should forward or backward to targetDate
   * @return an array of date
   */
  public static Date[] getScheduleDate(Date reference, Date targetDate, Date[] toUpdate, long schedule, boolean forward) {
    Calendar nowCal = Calendar.getInstance();
    nowCal.setTime(targetDate);

    Calendar refCal = Calendar.getInstance();
    refCal.setTime(reference);

    int[] modifiers = getScheduleModifier(schedule, forward);
    int loop = 0;
    if (!reference.equals(targetDate))
      loop = loopScheduleDate(refCal, nowCal, modifiers, forward);

    if (toUpdate == null)
      return null;

    Date[] toUpdateNew = new Date[toUpdate.length];
    for (int i = 0; i < toUpdate.length; i++) {

      Calendar toUpdateCal = Calendar.getInstance();
      toUpdateCal.setTime(toUpdate[i]);
      loopScheduleDate(toUpdateCal, modifiers, forward, loop);
      toUpdateNew[i] = toUpdateCal.getTime();
    }

    return toUpdateNew;
  }

  /**
   * Loop the reference calendar
   * 
   * @param refCal the reference starting calendar
   * @param nowCal the target calendar
   * @param modifiers the modifier provides by getScheduleModifier()
   * @param forward should forward or backward to targetDate
   * @return the number of iteration
   */
  public static int loopScheduleDate(Calendar refCal, Calendar nowCal, int[] modifiers, boolean forward) {
    return loopScheduleDate(refCal, nowCal, modifiers, forward, -1);
  }

  /**
   * Loop the reference calendar
   * 
   * @param refCal the reference starting calendar
   * @param modifiers the modifier provides by getScheduleModifier()
   * @param forward should forward or backward to targetDate
   * @param maxloop the loop to do or -1
   * @return the number of iteration
   */
  public static int loopScheduleDate(Calendar refCal, int[] modifiers, boolean forward, int maxloop) {
    return loopScheduleDate(refCal, null, modifiers, forward, maxloop);
  }

  /**
   * Loop the reference calendar
   * 
   * @param refCal the reference starting calendar
   * @param nowCal the target calendar
   * @param modifiers the modifier provides by getScheduleModifier()
   * @param forward should forward or backward to targetDate
   * @param maxloop the loop to do or -1
   * @return the number of iteration
   */
  private static int loopScheduleDate(Calendar refCal, Calendar nowCal, int[] modifiers, boolean forward, int maxloop) {

    if (maxloop < 0)
      maxloop = Integer.MAX_VALUE;

    Date time = refCal.getTime();
    Calendar tmp = Calendar.getInstance();
    tmp.setTime(time);

    for (int i = 0; i < maxloop; i++) {
      if (nowCal != null) {
        if (forward && tmp.after(nowCal))
          return i;
        if (!forward && tmp.before(nowCal))
          return i;
        if (i > 1 && tmp.equals(nowCal))
          return i;
      }

      tmp.setTime(time);

      // @formatter:off
      if (modifiers[6] != 0){ tmp.add(Calendar.MILLISECOND,  modifiers[6]*(i));}
      if (modifiers[5] != 0){ tmp.add(Calendar.SECOND,       modifiers[5]*(i));}
      if (modifiers[4] != 0){ tmp.add(Calendar.MINUTE,       modifiers[4]*(i));}
      if (modifiers[3] != 0){ tmp.add(Calendar.HOUR_OF_DAY,  modifiers[3]*(i));}
      if (modifiers[2] != 0){ tmp.add(Calendar.DAY_OF_WEEK,  modifiers[2]*(i));}
      if (modifiers[1] != 0){ tmp.add(Calendar.WEEK_OF_MONTH,modifiers[1]*(i));}
      if (modifiers[0] != 0){ tmp.add(Calendar.MONTH,        modifiers[0]*(i));}
      // @formatter:on

      refCal.setTime(tmp.getTime());
    }
    return maxloop;
  }

  /**
   * Parse schedule to return an array of modifiers for calendar roll
   * with the given values:
   * <br>
   * modifiers[0] month<br>
   * modifiers[1] weeks<br>
   * modifiers[2] days<br>
   * modifiers[3] hours<br>
   * modifiers[4] minutes<br>
   * modifiers[5] seconds<br>
   * modifiers[6] millis<br>
   * 
   * @param schedule the time to add to reference date.
   * @param forward should forward or backward to targetDate
   * @return an array of modifiers
   */
  public static int[] getScheduleModifier(long schedule, boolean forward) {

    int[] modifiers = new int[7];
    int dir = forward ? 1 : -1;
    long duration = schedule;

    // @formatter:off
    modifiers[0]  = dir * (int)(duration / JcmsConstants.MILLIS_IN_ONE_MONTH);   duration = duration % JcmsConstants.MILLIS_IN_ONE_MONTH;
    modifiers[1]  = dir * (int)(duration / JcmsConstants.MILLIS_IN_ONE_WEEK);    duration = duration % JcmsConstants.MILLIS_IN_ONE_WEEK;
    modifiers[2]  = dir * (int)(duration / JcmsConstants.MILLIS_IN_ONE_DAY);     duration = duration % JcmsConstants.MILLIS_IN_ONE_DAY;
    modifiers[3]  = dir * (int)(duration / JcmsConstants.MILLIS_IN_ONE_HOUR);    duration = duration % JcmsConstants.MILLIS_IN_ONE_HOUR;
    modifiers[4]  = dir * (int)(duration / JcmsConstants.MILLIS_IN_ONE_MINUTE);  duration = duration % JcmsConstants.MILLIS_IN_ONE_MINUTE;
    modifiers[5]  = dir * (int)(duration / JcmsConstants.MILLIS_IN_ONE_SECOND);  duration = duration % JcmsConstants.MILLIS_IN_ONE_SECOND;
    modifiers[6] = dir * (int) duration;
    // @formatter:on

    return modifiers;
  }

  /**
   * Parse the JCMS schedules
   * 
   * @param schedule the schedule
   * @param period the period ("everyhour", "everyday", "everyweek", "everymonth")
   * @param lang the language to format the result
   * @return a parsed schedule
   * @since jcms-5.7.0
   */
  public static String parseSchedule(String schedule, String period, String lang) {

    String output = "never";
    try {
      if (Util.notEmpty(schedule)) {
        DateFormat dfDate = channel.getDateTimeFormat(lang);
        Date date = dfDate.parse(schedule);

        Calendar current = Calendar.getInstance(I18nUtil.getLocale(lang));
        current.setTime(date);
        output = current.get(Calendar.MINUTE) + " ";

        if (period.equals("everyhour")) {
          output += "-1 -1 -1";
        } else if (period.equals("everyday")) {
          output += current.get(Calendar.HOUR_OF_DAY) + " -1 -1";
        } else if (period.equals("everyweek")) {
          output += current.get(Calendar.HOUR_OF_DAY) + " -1 " + current.get(Calendar.DAY_OF_WEEK);
        } else if (period.equals("everymonth")) {
          output += current.get(Calendar.HOUR_OF_DAY) + " " + current.get(Calendar.DAY_OF_MONTH) + " -1";
        } else {
          output = "never";
        }
      }
    } catch (Exception ex) {
      logger.warn("Bad Date format in JCMS schedule", ex);
    }
    return output;
  }

  /**
   * Format a cron-like schedule.
   * Example : schedule <code>30 20 * * 3 *</code> will be
   * formatted as <i>Every Tuesday at 8:30 PM</i>.
   * 
   * @param userLang the user language
   * @param schedule the cron schedule to be formatted.
   * @return a localized description of the schedule
   * @since jcms-5.5.0
   */
  public static String formatSchedule(String userLang, String schedule) {

    // Parse the schedule
    if (Util.isEmpty(schedule)) {
      return "";
    }

    StringTokenizer st = new StringTokenizer(schedule);
    String minuteStr = st.nextToken();
    String hourStr = st.nextToken();
    String dayOfMonthStr = st.nextToken();
    String monthStr = st.nextToken();
    String dayOfWeekStr = st.nextToken();
    // String yearStr = st.nextToken();

    int minute = "*".equals(minuteStr) ? 0 : Integer.parseInt(minuteStr);
    int hour = "*".equals(hourStr) ? 0 : Integer.parseInt(hourStr);
    int dayOfMonth = "*".equals(dayOfMonthStr) ? -1 : Integer.parseInt(dayOfMonthStr);
    int month = "*".equals(monthStr) ? -1 : Integer.parseInt(monthStr);
    int dayOfWeek = "*".equals(dayOfWeekStr) ? -1 : Integer.parseInt(dayOfWeekStr);
    // int year = "*".equals(yearStr) ? -1 : Integer.parseInt(yearStr);

    Locale locale = I18nUtil.getLocale(userLang);
    Calendar cal = Calendar.getInstance(locale);

    // Determine the type of schedule
    int EVERYDAY = 1;
    int WEEKLY = 2;
    int MONTHLY1 = 3;
    int MONTHLY2 = 4;

    int mode = EVERYDAY;
    cal.set(Calendar.MINUTE, minute);
    // to force update of the calendar object as precised in javadoc
    cal.getTime();
    cal.set(Calendar.HOUR_OF_DAY, hour);
    // to force update of the calendar object as precised in javadoc
    cal.getTime();

    // fix issues when current date is 31 and asked month doesn't have 31 days
    cal.set(Calendar.DAY_OF_MONTH, 1);
    cal.getTime();
    if (month != -1) {
      cal.set(Calendar.MONTH, month);
      // to force update of the calendar object as precised in javadoc
      cal.getTime();
      mode = MONTHLY2;
    }
    if (dayOfMonth != -1) {
      cal.set(Calendar.DAY_OF_MONTH, dayOfMonth);
      // to force update of the calendar object as precised in javadoc
      cal.getTime();
      if (mode == EVERYDAY) {
        mode = MONTHLY1;
      }
    }
    if (dayOfWeek != -1 && dayOfMonth == -1) {
      cal.set(Calendar.DAY_OF_WEEK, dayOfWeek);
      // to force update of the calendar object as precised in javadoc
      cal.getTime();
      mode = WEEKLY;
    }

    Date date = cal.getTime();

    // Format
    String timeStr = DateFormat.getTimeInstance(DateFormat.SHORT, locale).format(date);
    if (mode == WEEKLY) {
      String nameOfDay = (new SimpleDateFormat("EEEEE", locale)).format(date);
      return glp(userLang, "ui.schedule.weekly", nameOfDay, timeStr);
    }

    if (mode == MONTHLY1) {
      return glp(userLang, "ui.schedule.monthly1", "" + dayOfMonth, timeStr);
    }

    if (mode == MONTHLY2) {
      String nameOfMonth = (new SimpleDateFormat("MMMMM", locale)).format(date);
      return glp(userLang, "ui.schedule.monthly2", "" + dayOfMonth, nameOfMonth, timeStr);
    }

    return glp(userLang, "ui.schedule.everyday", timeStr);
  }

  /**
   * Create a new SortedSet of Date from a comma seperated time in millisecond.
   * 
   * @param timesString a String with comma sperated times in millissecond
   * @return a new SortedSet, never return null.
   * @since jcms-6.1.0
   */
  public static SortedSet<Date> stringToDateSet(String timesString) {
    List<String> timesStrList = Util.splitToList(timesString, ",");
    TreeSet<Date> dateSet = new TreeSet<Date>(ComparatorUtils.reversedComparator(ComparatorUtils.NATURAL_COMPARATOR));
    if (Util.notEmpty(timesStrList)) {
      for (String timeStr : timesStrList) {
        long time = Util.toLong(timeStr, 0);
        if (time > 0) {
          dateSet.add(new Date(time));
        }
      }
    }
    return dateSet;
  }

  /**
   * Create a new List of String time in millis from a SortedSet of Date.
   * 
   * @param dateSet a SortedSet of Date.
   * @return a new String, never return null.
   * @since jcms-6.1.0
   */
  public static String dateSetToString(SortedSet<Date> dateSet) {
    if (Util.isEmpty(dateSet)) {
      return "";
    }

    List<String> timeStrList = new ArrayList<String>(dateSet.size());
    for (Date itDate : dateSet) {
      timeStrList.add(String.valueOf(itDate.getTime()));
    }
    return Util.join(timeStrList, ",");
  }

  // ---------------------------------------------------------------------------------------------------------------
  // CORE Java Utility Method |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
  // ---------------------------------------------------------------------------------------------------------------

  /**
   * Return the Classpath of the webapp
   * 
   * @return String
   */
  public static String getWebappClasspath() {
    StringBuffer classpathSB = new StringBuffer();
    try {
      // AddchannelEB-INF/lib/
      String[] templJars = (new File(channel.getWebInfPath("templates/WEB-INF/lib"))).list();
      if (templJars != null) {
        for (int i = 0; i < templJars.length; i++) {
          classpathSB.append(channel.getWebInfPath("templates/WEB-INF/lib/" + templJars[i]) + System.getProperty("path.separator"));
        }
      }

      classpathSB.append(channel.getWebInfPath("classes/") + System.getProperty("path.separator"));
      File[] jars = (new File(channel.getWebInfPath("lib"))).listFiles(new RegexpFileFilter(".+\\.jar"));
      if (jars != null) {
        for (int i = 0; i < jars.length; i++) {
          classpathSB.append(channel.getWebInfPath("lib/" + jars[i].getName()) + System.getProperty("path.separator"));
        }
      }
      classpathSB.append(System.getProperty("java.class.path"));
    } catch (MalformedPatternException ex) {
      logger.warn("An exception occured while retrieving WebApp's classpath", ex);
    }
    return classpathSB.toString();
  }

  // ----------------------------------------------------------------
  // Ignored Resources FileFilter
  // ----------------------------------------------------------------

  /**
   * Retrieves a new instance of FileFilter used to <b>skip</b> resources that
   * should be ignored, as defined by property <code>"file.regexp.ignore"</code>.
   * <p>
   * Current implementation uses a reversed file filter of a regexp file filer.
   * <b>If the regexp defined by the property could not be compiled, a warning message is
   * logged and a TrueFileFilter (accepting all files) is returned.</b>
   * 
   * @return a new instance of FileFilter, never return null.
   * @since jcms-5.7.3
   * @see #getIgnoredResourcesFileFiler(boolean)
   */
  public static FileFilter getIgnoredResourcesFileFiler() {
    return getIgnoredResourcesFileFiler(false);
  }

  /**
   * Retrieves a new instance of FileFilter to either refuse or accept resources that
   * should be ignored, as defined by property <code>"file.regexp.ignore"</code>.
   * <p>
   * If accept parameter is set to false, uses a reversed file filter of a regexp file filer.
   * <br>
   * <b>If the regexp defined by the property could not be compiled, a warning message is
   * logged and a TrueFileFilter (accepting all files) is returned.</b>
   * 
   * @param accept true to accept ignored Files and refuse all other,
   *          false to refuse ignored Files and accept all others.
   * @return a new instance of FileFilter, never return null.
   * @since jcms-5.7.3
   */
  public static FileFilter getIgnoredResourcesFileFiler(boolean accept) {
    FileFilter filter = TrueFileFilter.INSTANCE;
    String regexp = channel.getProperty("file.regexp.ignore");
    try {
      filter = new RegexpFileFilter(regexp);
      if (!accept) {
        filter = new ReverseFileFilter(filter);
      }
    } catch (MalformedPatternException ex) {
      logger.warn("RegexpFileFilter could not be instanciated from property file.regexp.ignore '" + regexp + "'", ex);
    }
    return filter;
  }

  // ---------------------------------------------------------------------------------------------------------------
  // Generic String Replace |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
  // ---------------------------------------------------------------------------------------------------------------

  public static final Pattern DUPLICATE_PREFIX;
  static {
    Pattern tmpPattern = null;
    try {
      PatternCompiler compiler = new Perl5Compiler();
      tmpPattern = compiler.compile("\\$NAME\\$");
    } catch (MalformedPatternException ex) {
      logger.warn("An exception occured while compiling duplicate perl5 pattern", ex);
    }
    DUPLICATE_PREFIX = tmpPattern;
  }

  /**
   * Construct a String with prefix + title or replace a $NAME$ by this prefix
   * 
   * @param prefix the prefix to append
   * @param title the title
   * @param replaceOnly if true does not append the prefix
   * @return the String correctly formated
   * @since jcms-4.1
   */
  public static String getPrefixedTitle(String prefix, String title, boolean replaceOnly) {

    if (Util.isEmpty(title))
      return null;

    if (Util.isEmpty(prefix))
      return title;

    Perl5Matcher matcher = new Perl5Matcher();
    if (matcher.contains(title, DUPLICATE_PREFIX)) {
      Perl5Substitution substitute = new Perl5Substitution(prefix);
      return org.apache.oro.text.regex.Util.substitute(matcher, DUPLICATE_PREFIX, substitute, title, org.apache.oro.text.regex.Util.SUBSTITUTE_ALL);
    }
    return replaceOnly ? title : (prefix + title);
  }

  // ---------------------------------------------------------------------------------------------------------------
  // JSP to File or to Output Stream |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
  // ---------------------------------------------------------------------------------------------------------------

  /**
   * Render a JSP into a File.
   * 
   * @param jsp the JSP to be rendered. Path relative to the webapp root folder, must start with "/". E.g :
   *          <code>/plugins/FooBarPlugin/jsp/dump.jsp</code>.
   * @param filename the output file where the jsp process output is written. Path relative to the webapp root folder, with or without starting slash.
   *          E.g. : <code>/WEB-INF/tmp/dump.txt</code>.
   * @param sessionAttributeMap the session attributes
   * @throws Exception If an error occurs
   */
  public static void jsp2File(String jsp, String filename, HashMap sessionAttributeMap) throws Exception {

    // @formatter:off
    jsp2File(
        jsp, 
        filename, 
        null, 
        null, 
        sessionAttributeMap, 
        null, 
        null, 
        null);
    // @formatter:on
  }

  /**
   * Render a JSP into a File.
   * 
   * @param jsp the JSP to be rendered. Path relative to the webapp root folder, must start with "/". E.g :
   *          <code>/plugins/FooBarPlugin/jsp/dump.jsp</code>.
   * @param filename the output file where the jsp process output is written. Path relative to the webapp root folder, with or without starting slash.
   *          E.g. : <code>/WEB-INF/tmp/dump.txt</code>.
   * @param requestParameterMap the request parameters
   * @param requestAttributeMap the request attributes
   * @param sessionAttributeMap the session attributes
   * @param contextAttributeMap the context attributes
   * @param initParameterMap the init parameters
   * @param loggedMember the logged member used to process the JSP
   * @throws Exception If an error occurs
   */
  // @formatter:off
  public static void jsp2File(
      String jsp,
      String filename,
      HashMap requestParameterMap,
      HashMap requestAttributeMap,
      HashMap sessionAttributeMap,
      HashMap contextAttributeMap,
      HashMap initParameterMap,
      Member loggedMember) throws Exception {

    jsp2File(
        jsp,
        filename,
        requestParameterMap,
        requestAttributeMap,
        sessionAttributeMap,
        contextAttributeMap, 
        initParameterMap,
        loggedMember,
        false);
  }
  // @formatter:on

  /**
   * Render a JSP into a File.
   * 
   * @param jsp the JSP to be rendered. Path relative to the webapp root folder, must start with "/". E.g :
   *          <code>/plugins/FooBarPlugin/jsp/dump.jsp</code>.
   * @param filename the output file where the jsp process output is written. Path relative to the webapp root folder, with or without starting slash.
   *          E.g. : <code>/WEB-INF/tmp/dump.txt</code>.
   * @param requestParameterMap the request parameters
   * @param requestAttributeMap the request attributes
   * @param sessionAttributeMap the session attributes
   * @param contextAttributeMap the context attributes
   * @param initParameterMap the init parameters
   * @param loggedMember the logged member used to process the JSP
   * @param clean if true the compiled servlet is removed once processed.
   * @since jcms-4.0
   * @throws Exception If an error occurs
   */
  // @formatter:off
  public static void jsp2File(
      String jsp,
      String filename,
      HashMap requestParameterMap,
      HashMap requestAttributeMap,
      HashMap sessionAttributeMap,
      HashMap contextAttributeMap,
      HashMap initParameterMap,
      Member loggedMember,
      boolean clean) throws Exception {

    BufferedOutputStream bos = null;
    try {
      File file = new File(channel.getRealPath(filename));
      bos = new BufferedOutputStream(new FileOutputStream(file));
      jsp2OutputStream(
                       jsp,
                       bos,
                       requestParameterMap,
                       requestAttributeMap,
                       sessionAttributeMap,
                       contextAttributeMap,
                       initParameterMap,
                       loggedMember,
          clean);
      bos.flush();
    } finally {
      IOUtil.closeQuietly(bos);
    }
  }
  // @formatter:on

  /**
   * Render a JSP into an OutputStream
   * 
   * @param jsp the JSP to be rendered. It must start with "/".
   * @param out the outputstream
   * @param sessionAttributeMap the session attributes
   * @throws Exception If an error occurs
   */
  // @formatter:off
  public static void jsp2OutputStream(
      String jsp,
      OutputStream out,
      HashMap sessionAttributeMap) throws Exception {

    jsp2OutputStream(
        jsp,
        out,
        null,
        null,
        sessionAttributeMap,
        null,
        null,
        null);
  }
  // @formatter:on

  /**
   * Render a JSP into an OutputStream
   * 
   * @param jsp the JSP to be rendered. It must start with "/".
   * @param out the outputstream
   * @param requestParameterMap the request parameters
   * @param requestAttributeMap the request attributes
   * @param sessionAttributeMap the session attributes
   * @param contextAttributeMap the context attributes
   * @param initParameterMap the init parameters
   * @param loggedMember the logged member used to process the JSP
   * @throws Exception If an error occurs
   */
  // @formatter:off
  public static void jsp2OutputStream(
      String jsp,
      OutputStream out,
      HashMap requestParameterMap,
      HashMap requestAttributeMap,
      HashMap sessionAttributeMap,
      HashMap contextAttributeMap,
      HashMap initParameterMap,
      Member loggedMember) throws Exception {

    jsp2OutputStream(
        jsp,
        out,
        requestParameterMap,
        requestAttributeMap,
        sessionAttributeMap,
        contextAttributeMap,
        initParameterMap,
        loggedMember,
        false);
  }
  // @formatter:on

  /**
   * Render a JSP into an OutputStream
   * 
   * @param jsp the JSP to be rendered. It must start with "/".
   * @param out the outputstream
   * @param requestParameterMap the request parameters
   * @param requestAttributeMap the request attributes
   * @param sessionAttributeMap the session attributes
   * @param contextAttributeMap the context attributes
   * @param initParameterMap the init parameters
   * @param loggedMember the logged member used to process the JSP
   * @param clean if true the compiled servlet is removed once processed.
   * @throws Exception If an error occurs
   * @since jcms-4.0
   */
  // @formatter:off
  public static void jsp2OutputStream(
      String jsp,
      OutputStream out,
      HashMap requestParameterMap,
      HashMap requestAttributeMap,
      HashMap sessionAttributeMap,
      HashMap contextAttributeMap,
      HashMap initParameterMap,
      Member loggedMember,
      boolean clean) throws Exception {

    // Context attributes
    if (contextAttributeMap == null) {
      contextAttributeMap = new HashMap();
    }
    contextAttributeMap.put("channel", channel);

    // Request attributes
    if (requestAttributeMap == null) {
      requestAttributeMap = new HashMap();
    }
    String userLang = null;
    if (loggedMember != null) {
      requestAttributeMap.put("loggedMember", loggedMember);
      // Check user lang is a valid channel language
      if (channel.getLanguageList().contains(loggedMember.getLanguage())) {
        userLang = loggedMember.getLanguage();
      }
    }
    // make sure we always define a userLang and a userLocale for doInitMember
    if (requestAttributeMap.get("userLang") == null) {
      requestAttributeMap.put("userLang", userLang != null ? userLang : channel.getLanguage());
    }
    Locale userLocale = (Locale) requestAttributeMap.get("userLocale");
    if (userLocale == null) {
      userLang = (String) requestAttributeMap.get("userLang");
      userLocale = I18nUtil.getLocale(userLang);
      requestAttributeMap.put("userLocale", userLocale);
    }

    jsp2OutputStream(jsp,
                     out,
                     requestParameterMap,
                     requestAttributeMap,
                     sessionAttributeMap,
                     contextAttributeMap,
                     initParameterMap,
                     channel.getJcmsWorkPath("Trash"),
                     channel.getWebappPath(),
                     channel.getClass().getClassLoader(),
                     getWebappClasspath(),
                     clean);
  }
  // @formatter:on

  /**
   * Render a JSP into an OutputStream
   * 
   * @param jsp the JSP to be rendered. It must start with "/".
   * @param out the outputstream
   * @param requestParameterMap the request parameters
   * @param requestAttributeMap the request attributes
   * @param sessionAttributeMap the session attributes
   * @param contextAttributeMap the context attributes
   * @param initParameterMap the init parameters
   * @param trashPath the trash path
   * @param rootPath the root path
   * @param classloader the classloader
   * @param classpath the classpath
   * @throws Exception If an error occurs
   */
  // @formatter:off
  public static void jsp2OutputStream(
      String jsp,
      OutputStream out,
      HashMap requestParameterMap,
      HashMap requestAttributeMap,
      HashMap sessionAttributeMap,
      HashMap contextAttributeMap,
      HashMap initParameterMap,
      String trashPath,
      String rootPath,
      ClassLoader classloader,
      String classpath) throws Exception {
    jsp2OutputStream(
        jsp,
        out, 
        requestParameterMap,
        requestAttributeMap,
        sessionAttributeMap,
        contextAttributeMap,
        initParameterMap,
        trashPath,
        rootPath,
        classloader,
        classpath,
        false);
  }
  // @formatter:on

  /**
   * Render a JSP into an OutputStream
   * 
   * @param jsp the JSP to be rendered. It must start with "/".
   * @param out the outputstream
   * @param requestParameterMap the request parameters
   * @param requestAttributeMap the request attributes
   * @param sessionAttributeMap the session attributes
   * @param contextAttributeMap the context attributes
   * @param initParameterMap the init parameters
   * @param trashPath the trash path
   * @param rootPath the root path
   * @param classloader the classloader
   * @param classpath the classpath
   * @param clean if true the compiled servlet is removed once processed.
   * @throws Exception If an error occurs
   */
  // @formatter:off
  public static void jsp2OutputStream(
      String jsp,
      OutputStream out,
      HashMap requestParameterMap,
      HashMap requestAttributeMap,
      HashMap sessionAttributeMap,
      HashMap contextAttributeMap,
      HashMap initParameterMap,
      String trashPath,
      String rootPath,
      ClassLoader classloader,
      String classpath,
      boolean clean) throws Exception {

    File trash = new File(trashPath);
    if (trash.exists() && clean) {
      IOUtil.deepDelete(trash);
    }

    // WARNING : channel might be null when invoked by JSPtoFile.main()
    HttpServletRequest currentRequestBackup = channel != null ? channel.getCurrentServletRequest() : null;
    try {
      TemplateEngine engine = new TemplateEngine(rootPath, trash.getAbsolutePath(), classloader, classpath, "error");

      FakedServletContext context = new FakedServletContext(rootPath, engine, contextAttributeMap, initParameterMap);
      FakedHttpSession session = new FakedHttpSession(context, sessionAttributeMap);
      FakedServletRequest request = new FakedServletRequest(session, requestAttributeMap, requestParameterMap, jsp);
      FakedServletResponse response = new FakedServletResponse(out);

      // Force all jsp and template to be rendered as UTF-8
      response.setCharacterEncoding("UTF-8");

      if (channel != null) {
        channel.setCurrentServletRequest(request);
      }

      synchronized (jspEngineLock) {
        engine.render(request, response, jsp);
      }

    } finally {
      // restore
      if (channel != null) {
        if (currentRequestBackup == null) {
          channel.clearCurrentServletRequest();
        } else {
          channel.setCurrentServletRequest(currentRequestBackup);
        }
      }
    }
  }
  // @formatter:on

  /**
   * Convenient method that clean the generated files by template engine.
   * Use before multiple jsp2file ....
   */
  public static void jsp2FileCleanTrash() {
    File trash = new File(channel.getJcmsWorkPath("Trash"));
    if (trash.exists()) {
      IOUtil.deepDelete(trash);
    }
  }

  // ---------------------------------------------------------------------------------------------------------------
  // WebApp Utility Method ||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
  // ---------------------------------------------------------------------------------------------------------------

  public static void restartWebapp(File file) {
    if (!file.exists()) {
      try {
        PrintWriter pw = new PrintWriter(new FileWriter(file));
        pw.println("Created by JCMS. " + new Date());
        pw.println("This file must be present and be updated for channel restart.");
        pw.close();
      } catch (java.io.IOException ex) {
        logger.warn("An exception occured while restarting WebApp with " + file, ex);
      }
    }
    file.setLastModified(System.currentTimeMillis());
    try {
      Thread.sleep(2000);
    } catch (Exception ex) {
      logger.debug("Exception occured while waiting for webapp restart", ex);
    }
  }

  // ---------------------------------------------------------------------------------------------------------------
  // WYSIWYG Utility Method |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
  // ---------------------------------------------------------------------------------------------------------------

  /**
   * Replace "&lt;" and "&amp;" with their corresponding HTML entities
   * (&amp;lt; and &amp;amp;) in the specified String.
   * <p>
   * The side effect of using this method is that the resulting content can only be
   * outputed to an HTML stream. If the content is printed in a plaintext, csv, mail
   * or other non HTML stream, characters "&lt;" and "&amp;" will appear as HTML entity.<br/>
   * To limit this side effect, ampersand are only escaped when they would match
   * an HTML entity (thus being interpreted by browser at display/re-edition).
   * <p>
   * This method perform 2 pattern replacements :
   * 
   * <pre>
   *  s/ &amp;([^\s;]{2,20}); / &amp;amp$1; /g
   *  s/ &lt; / &amp;lt; /g
   * </pre>
   * 
   * Ampersand (&amp;) are converted in the corresponding HTML entity since
   * JCMS 7.0.4 and 7.1.1.<br/>
   * Prior to those version only the "&lt;" was being replaced.
   * See issue <a href="http://issues.jalios.com/browse/JCMS-2791">JCMS-2791</a> for more information on this matter.
   * 
   *
   * @param str the string to be escaped
   * @return the escaped string
   * @since jcms-5.0.0
   */
  public static String escapeHtml(String str) {
    if (Util.isEmpty(str)) {
      return str;
    }
    str = new Perl5Util().substitute("s/&([^\\s;]{2,20});/&amp;$1;/g", str); // Fix JCMS-2791
    str = new Perl5Util().substitute("s/</&lt;/g", str);
    return str;
  }

  /**
   * Replace HTML entities "&amp;lt;" and "&amp;amp;" with their corresponding charaters
   * (&lt; and &amp;) in the specified String.
   * <p>
   * This method does the opposite work of {@link #escapeHtml(String)}.
   *
   * @param str the string to be escaped
   * @return the escaped string
   * @since jcms-5.0.0
   */
  public static String unescapeHtml(String str) {
    if (Util.isEmpty(str)) {
      return str;
    }
    str = new Perl5Util().substitute("s/&amp;/&/g", str);
    str = new Perl5Util().substitute("s/&lt;/</g", str);
    return str;
  }

  /**
   * Check if the given String does not contain forbidden HTML
   * code. Checks with channel.forbidden-html regexp, then delegates
   * to custom.RightPolicy.checkHtml().
   *
   * @param str the string to be checked
   * @param field the name of the field which is checked
   * @return a ControllerStatus
   * @see com.jalios.jcms.policy.BasicRightPolicyFilter#checkHtml(String,String)
   * @since jcms-5.0.0
   */
  public static ControllerStatus checkHtml(String str, String field) {
    if (Util.isEmpty(str)) {
      return ControllerStatus.OK;
    }

    // Compile the pattern
    if (forbiddenNoPattern) {
      return ControllerStatus.OK;
    }

    if (forbiddenPattern == null) {
      String regexp = channel.getProperty("channel.forbidden-html", "");
      if (Util.isEmpty(regexp)) {
        logger.info("No pattern to check forbidden HTML");
        forbiddenNoPattern = true;
        return ControllerStatus.OK;
      }
      try {
        PatternCompiler perl5Compiler = new Perl5Compiler();
        forbiddenPattern = perl5Compiler.compile(regexp, Perl5Compiler.CASE_INSENSITIVE_MASK);
        logger.info("Enable forbidden HTML pattern");
      } catch (Exception ex) {
        logger.warn("An exception occured while compiling channel.forbidden-html pattern: " + regexp, ex);
        forbiddenNoPattern = true;
        return ControllerStatus.OK;
      }
    }

    // Check the pattern
    PatternMatcher perl5Matcher = new Perl5Matcher();
    if (perl5Matcher.contains(str, forbiddenPattern)) {
      String match = perl5Matcher.getMatch().toString();
      logSecurityIssue(logger, Level.INFO, "Reject HTML code. It contains '" + match + "'");
      match = escapeHtml(match);
      ControllerStatus status = new ControllerStatus();
      if (Util.isEmpty(field)) {
        status.setProp("msg.edit.forbidden-html-code", match);
      } else {
        status.setProp("msg.edit.forbidden-html-field", match, field);
      }
      return status;
    }

    // Call RightPolicy
    return channel.getPolicyManager().checkHtml(str, field);
  }

  /**
   * Check if the given array of String does not contain forbidden HTML code
   *
   * @param array the array of string to be checked
   * @param field the name of the field which is checked
   * @return a ControllerStatus
   * @see #checkHtml(String,String)
   * @since jcms-5.0.0
   */
  public static ControllerStatus checkHtml(String[] array, String field) {
    if (array == null) {
      return ControllerStatus.OK;
    }
    for (int i = 0; i < array.length; i++) {
      ControllerStatus status = checkHtml(array[i], field);
      if (status.hasFailed()) {
        return status;
      }
    }
    return ControllerStatus.OK;
  }

  /**
   * Check if the values of given map of String do not contain forbidden HTML code
   *
   * @param map the map with the values to be checked
   * @param field the name of the field which is checked
   * @return a ControllerStatus
   * @see #checkHtml(String,String)
   * @since jcms-5.0.0
   */
  public static ControllerStatus checkHtml(Map map, String field) {
    if (map == null || map.values() == null) {
      return ControllerStatus.OK;
    }
    for (Iterator it = map.values().iterator(); it.hasNext();) {
      Object obj = it.next();
      if (obj instanceof String) {
        ControllerStatus status = checkHtml((String) obj, field);
        if (status.hasFailed()) {
          return status;
        }
      }
      if (obj instanceof String[]) {
        ControllerStatus status = checkHtml((String[]) obj, field);
        if (status.hasFailed()) {
          return status;
        }
      }
    }
    return ControllerStatus.OK;
  }

  // ---------------------------------------------------------------------------------------------------------------
  // Data Utility Method ||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
  // ---------------------------------------------------------------------------------------------------------------

  /**
   * Convert an old array of Data to a new array of data using a Map
   * 
   * @param old an array of Data
   * @param dataMap the data map to fill
   * @return a Data[] updated
   */
  public static Data[] getUpdatedData(Data[] old, Map dataMap) {
    ArrayList newList = new ArrayList();

    if (old == null)
      return null;

    if (dataMap == null)
      return old;

    for (int i = 0; i < old.length; i++) {
      Data itData = old[i];
      if (itData == null)
        continue;
      Data newData = (Data) dataMap.get(itData);
      if (newData == null)
        continue;
      newList.add(newData);
    }
    return (Data[]) newList.toArray(old);
  }

  /**
   * Convert an old set of Data to a new Set of data using a Map
   * 
   * @param old the old set of data
   * @param dataMap the map
   * @param retrieveAll if newdata doesn't exists then leave old data
   * @return a Set of Data updated
   */
  public static TreeSet getUpdatedData(Collection old, Map dataMap, boolean retrieveAll) {
    TreeSet newSet = new TreeSet();

    if (old == null)
      return newSet;

    if (dataMap == null) {
      if (retrieveAll)
        newSet.addAll(old);
      return newSet;
    }

    for (Iterator it = old.iterator(); it.hasNext();) {
      Data itData = (Data) it.next();
      if (itData == null)
        continue;
      Data newData = (Data) dataMap.get(itData);
      if (newData == null) { // If data doesn't exists then add it
        if (retrieveAll) {
          newSet.add(itData);
        }
      } else {
        newSet.add(newData);
      }
    }
    return newSet;
  }

  /**
   * Convert an old set of Data to a new Set of data using a Map
   * 
   * @param old the old set of data
   * @param dataMap the map
   * @return a Set of Data updated
   * @see #getUpdatedData(Collection, Map, boolean)
   */
  public static TreeSet getUpdatedData(Collection old, Map dataMap) {
    return getUpdatedData(old, dataMap, false);
  }

  // -----------------------------------------------
  // Data DCM Utilities methods
  // -----------------------------------------------

  /**
   * This convenient method checks if the given multilingual field is required.
   * 
   * @param pub the publication
   * @param fieldName the name of the field
   * @param value the value of the field
   * @param valueML the ML value of the field
   * @return a ControllerStatus
   * @since jcms-6.1.2
   */
  public static ControllerStatus checkFieldMLRequired(Publication pub, String fieldName, String value, HashMap valueML) {
    if (Util.notEmpty(pub.getFieldInMainLanguage(value, valueML))) {
      return ControllerStatus.OK;
    }

    ControllerStatus status = new ControllerStatus();
    status.setProp("msg.edit.empty-fieldml", channel.getTypeFieldLabel(pub, fieldName), new LangProperty("lang." + pub.getMainLanguage()));
    return status;
  }

  /**
   * This convenient method checks if the given field is required.
   * 
   * @param pub the publication
   * @param fieldName the name of the field
   * @param value the value of the field
   * @return a ControllerStatus
   * @since jcms-6.1.2
   */
  public static ControllerStatus checkFieldRequired(Publication pub, String fieldName, Object value) {
    if (Util.notEmpty(value)) {
      return ControllerStatus.OK;
    }

    ControllerStatus status = new ControllerStatus();
    status.setProp("msg.edit.empty-field", channel.getTypeFieldLabel(pub, fieldName));
    return status;
  }

  // -----------------------------------------------
  // Language Properties
  // -----------------------------------------------

  /**
   * Resolve a language property in the default site language.<br>
   * <b>Warning</b>: You should always prefer the use of 'glp' methods receiving a
   * language parameter in order to resolve the property in the language of the
   * end user.
   * 
   * @param prop the language property to resolve, for example <tt>"ui.com.lbl.contents"</tt>
   * @since jcms-5.5.0
   * @return the resolved property, or the property's name if its value could not
   *         be found in the specified language or in the default language
   */
  public static String glpd(String prop) {
    return JcmsUtil.glp(channel.getLanguage(), prop);
  }

  /**
   * Resolve a language property in the default site language, defaulting
   * to a null string if the property's value could not be found (instead of
   * defaulting to the property's name which is the default behavior when accessing
   * properties).<br>
   * <b>Warning</b>: You should always prefer the use of 'glp' methods receiving a
   * language parameter in order to resolve the property in the language of the
   * end user.
   * 
   * @param prop the language property to resolve, for example <tt>"ui.com.lbl.contents"</tt>
   * @since jcms-8.0.0
   * @return the resolved property
   */
  public static String glpdNoDefault(String prop) {
    String value = JcmsUtil.glp(channel.getLanguage(), prop);
    if (value != null && value.equals(prop)) {
      return null;
    }
    return value;
  }

  /**
   * Resolve a language property in the given language.
   * <p>
   * Example :
   * 
   * <pre>
   * assertEquals("1 byte", JcmsUtil.glpNull("en", "ui.com.lbl.byte", 1));
   * assertEquals("6 bytes", JcmsUtil.glpNull("en", "ui.com.lbl.byte", 6));
   * assertEquals("non.existing.property", JcmsUtil.glpNull("en", "non.existing.property"));
   * </pre>
   * 
   * @param lang the language in which you want the given property
   * @param prop the language property to resolve, for example <tt>"ui.com.lbl.contents"</tt>
   * @param params an array of parameters to use to for the format of property
   * @return the resolved property, or the property's name if its value could not
   *         be found in the specified language or in the default language
   * @since jcms-5.0.0
   */
  public static String glp(String lang, String prop, Object... params) {
    return channel.getLanguageProperties().getExplicitLangProperty(prop, lang, params);
  }

  /**
   * Resolve a language property in the given language, using named parameters.
   * <p>
   * Example :
   * 
   * <pre>
   * Map<String, String> paramMap = new HashMap<String, String>();
   * paramMap.put("firstName", "John");
   * paramMap.put("lastName", "Doe");
   * String i18nStr = getExplicitLangProperty("my.property", "en", paramMap);
   * </pre>
   * 
   * with the given property:
   * 
   * <pre>
   * my.property: Hello {firstName} {lastName}
   * </pre>
   * 
   * @param lang the language in which you want the given property
   * @param prop the language property to resolve, for example <tt>"ui.com.lbl.contents"</tt>
   * @param paramMap an map of parameters, where the key is the name of the parameter in the property
   * @return the resolved property, or the property's name if its value could not
   *         be found in the specified language or in the default language
   */
  public static String glp(String lang, String prop, Map<String, String> paramMap) {
    return channel.getLanguageProperties().getExplicitLangProperty(prop, lang, paramMap);
  }

  /**
   * Resolve a language property in the given language.
   * 
   * @param lang the language in which you want the given property
   * @param prop the {@link LangProperty} instance corresponding to the property you want to resolve.
   * @return the resolved property
   * @since jcms-5.5.0
   */
  public static String glp(String lang, LangProperty prop) {
    return channel.getLanguageProperties().getExplicitLangProperty(prop, lang);
  }

  /**
   * Resolve a language property in the given language, defaulting
   * to a null value if the property's value could not be found (instead of
   * defaulting to the property's name which is the default behavior when accessing
   * properties).<br>
   * <p>
   * Example :
   * 
   * <pre>
   * assertEquals("1 byte", JcmsUtil.glpNull("en", "ui.com.lbl.byte", 1));
   * assertEquals("6 bytes", JcmsUtil.glpNull("en", "ui.com.lbl.byte", 6));
   * assertNull(JcmsUtil.glpNull("en", "non.existing.property"));
   * </pre>
   * 
   * @param lang the language in which you want the given property
   * @param prop the language property to resolve, for example <tt>"ui.com.lbl.contents"</tt>
   * @param params an array of parameters to use to for the format of property
   * @return the resolved property
   * @since jcms-8.0.0
   */
  public static String glpNoDefault(String lang, String prop, Object... params) {
    String value = channel.getLanguageProperties().getExplicitLangProperty(prop, lang, params);
    if (value != null && value.equals(prop)) {
      return null;
    }
    return value;
  }

  // -----------------------------------------------
  // Log4j MDC and NDC
  // -----------------------------------------------

  /**
   * Initialize log4j {@link org.apache.log4j.MDC} variables for the current thread.
   * <li>{ChannelName} => current site name</li>.
   * <p><strong>Important</strong> Do not use this method anymore as it cannot
   * be easily guaranteed to be memory leak free.
   * 
   * @deprecated uses of log4j MDC cannot easily be guaranteed without memory leak
   *             instead, use {@link JcmsPatternLayout} in log4j.xml with the pattern <code>%s</code>
   */
  public static void setLog4jMDC() {
    org.apache.log4j.MDC.put("ChannelName", (channel != null) ? channel.getName() : "JCMS");
  }

  /**
   * Initialize log4j {@link org.apache.log4j.NDC} with the given
   * logged member information. <br>
   * This methods also logs the current delegate Member when
   * available ({@link Channel#getCurrentDelegateMember()}].
   * 
   * @param request the current HttpServletRequest
   * @param loggedMember the Member currently logged
   */
  public static void setLog4jNDCMemberInfo(HttpServletRequest request, Member loggedMember) {
    final String infos;
    if (loggedMember == null) {
      infos = "[" + request.getRemoteAddr() + ", unauthenticated user]";
    } else {
      Member delegateMember = (channel != null) ? channel.getCurrentDelegateMember() : null;
      if (delegateMember != null) {
        // There is a delegate connected as somebody else
        // @formatter:off
        infos = "[" + request.getRemoteAddr() + ", " 
                + delegateMember.getId() + "/" + loggedMember.getId() + ", "
                + delegateMember + " / " + loggedMember + "]";
        // @formatter:on
      } else {
        // @formatter:off
        infos = "[" + request.getRemoteAddr() + ", "
                + loggedMember.getId() + ", "
                + loggedMember + "]";
        // @formatter:on
      }
    }

    // Remove all NDC currently set, but keep them in a stack to push
    // them back (all but the first one which is the old logged member)
    Stack ndcStack = new Stack();
    for (String ndc = NDC.pop(); Util.notEmpty(ndc); ndc = NDC.pop()) {
      ndcStack.push(ndc);
    }

    // Push new NDC on first pos
    NDC.push(infos);

    // Push back old NDC, (excepted first one)
    for (int i = 0; ndcStack.size() > 0; i++) {
      String ndc = (String) ndcStack.pop();
      if (i == 0) {
        continue;
      }
      NDC.push(ndc);
    }
  }

  // -----------------------------------------------
  // Log4j Forbidden and security issue
  // -----------------------------------------------

  /**
   * Log a forbidden access message (using the logSecurityIssue method)
   * A forbidden acces to the jsp http://www.mysite.com/admin.jsp,
   * with a pattern layout "%x %m", would output:
   * [SECURITY] Forbidden access to http://www.mysite.com/admin.jsp
   * 
   * @param lgr the logger to use
   * @param level the level of the log message
   * @param request the request where the forbidden is issued from
   * @since jcms-5.0.0
   */
  public static void logForbiddenAccess(Logger lgr, Level level, HttpServletRequest request) {
    logSecurityIssue(lgr, level, "Forbidden access to " + ServletUtil.getUrl(request));
  }

  /**
   * Log a forbidden message to the given request, at the WARN level using the given logger.
   * 
   * @param lgr the Logger
   * @param request the request
   * @see #logForbiddenAccess(Logger lgr, Level level, HttpServletRequest request)
   */
  public static void logForbiddenAccess(Logger lgr, HttpServletRequest request) {
    logForbiddenAccess(lgr, Level.WARN, request);
  }

  /**
   * Log a forbidden message to the given request, at the WARN level using the JcmsUtil logger.
   * 
   * @param request the request
   * @see #logForbiddenAccess(Logger lgr, Level level, HttpServletRequest request)
   */
  public static void logForbiddenAccess(HttpServletRequest request) {
    logForbiddenAccess(logger, request);
  }

  /**
   * Log the given message (object), at a the given level,
   * with the constant JcmsConstants.LOG_TOPIC_SECURITY as a log4j NDC using the
   * given logger.
   * 
   * The output with a pattern layout "%x %m" would be :
   * [SECURITY] message
   *
   * @param lgr the logger to use
   * @param level the level of the log message
   * @param obj the object (message) to log
   * @since jcms-5.0.0
   */
  public static void logSecurityIssue(Logger lgr, Level level, Object obj) {
    if (logger == null || level == null) {
      return;
    }
    NDC.push(JcmsConstants.LOG_TOPIC_SECURITY);
    lgr.log(level, obj);
    NDC.pop();
  }

  /**
   * Log the given message (object), at a the WARN level using the given logger.
   * 
   * @param lgr the Logger
   * @param obj the object
   * @see #logSecurityIssue(Logger lgr, Level level, Object obj)
   */
  public static void logSecurityIssue(Logger lgr, Object obj) {
    logSecurityIssue(lgr, Level.WARN, obj);
  }

  /**
   * Log the given message (object), at a the WARN level using the JcmsUtil logger.
   * 
   * @param obj the object
   * @see #logSecurityIssue(Logger lgr, Level level, Object obj)
   */
  public static void logSecurityIssue(Object obj) {
    logSecurityIssue(logger, obj);
  }

  // -----------------------------------------------
  // Pub icon and image
  // -----------------------------------------------

  /**
   * Returns the image of the icon of the given publication
   * 
   * @param pub the publication
   * @return the relative path of the icon or null if no icon
   * @since jcms-5.0.1
   */
  public static String getPubImg(Publication pub) {
    if (pub == null) {
      return null;
    }

    String img = null;

    if (pub instanceof FileDocument) {
      img = ((FileDocument) pub).getIcon();
    }

    if (pub instanceof PortalElement) {
      img = ((PortalElement) pub).getIcon();
    }

    return img;
  }

  /**
   * Returns the HTML code for the icon of the given publication
   * 
   * @param pub the publication
   * @return the html code of the image or null if no icon
   * @since jcms-5.0.1
   */
  public static String getPubIcon(Publication pub) {

    if (pub == null) {
      return null;
    }

    String img = getPubImg(pub);
    if (img == null) {
      return null;
    }

    String res = JcmsUtil.getHtmlIcon(img, null);
    if (pub instanceof FileDocument) {
      res = "<a href='work/displayWork.jsp?id=" + pub.getId() + "&amp;details=true'>" + res + "</a>";
    }
    return res;
  }

  /**
   * Gets the icon source from a {@link Class} object.
   *
   * @param clazz
   *          {@link Class}
   * @return String representing the icon source
   * @since jcms-8.0.0
   */
  public static String getClassIcon(Class<?> clazz) {
    if (clazz == null) {
      return null;
    }

    if (ICON_CACHE.containsKey(clazz)) {
      return ICON_CACHE.get(clazz);
    }
    String src = null;
    Class<?> c = clazz;
    while (Util.isEmpty(src) && c != null && Data.class.isAssignableFrom(c)) {
      src = channel.getProperty(JcmsConstants.TYPES_PREFIX_PROP + "" + c.getSimpleName() + "" + JcmsConstants.TYPES_ICON_SUFFIX_PROP, "");
      c = c.getSuperclass();
    }
    if (Util.notEmpty(src)) {
      ICON_CACHE.put(clazz, src);
    }
    return src;
  }

  /**
   * Build an icon from a {@link Class} object.
   *
   * @param clazz
   *          {@link Class}
   * @return the image tag representation
   * @since jcms-8.0.0
   */
  public static String getHtmlIcon(Class<?> clazz) {
    return HtmlUtil.getHtmlIcon(getClassIcon(clazz), "");
  }

  /**
   * Check if the given data is not null and stored
   * 
   * @param data the data to be checked
   * @return true if data is not null and stored.
   * @since jcms-5.0.2
   */
  public static boolean isStored(Data data) {
    return data != null && data.isStored();
  }

  /**
   * Check if the given data has been persisted (either in JStore or in the database).
   * 
   * @param data the data to be checked
   * @return true if data is not null and has been persisted.
   * @since jcms-6.1.0
   */
  public static boolean isPersisted(Data data) {
    return data != null && data.isPersisted();
  }

  // -----------------------------------------------
  // Selection methods
  // -----------------------------------------------

  /**
   * Perform a selection on a <code>Collection</code> of <code>Data</code> using
   * the given <code>DataSelector</code>, and return a TreeSet sorted using the
   * given <code>Comparator</code>. <br>
   * If no <code>Comparator</code> is given, <code>CdateComparator</code> is used. <br>
   *
   * @param <T> any Data type
   * @param dataColl the <code>Collection</code> of <code>Data</code> to select.
   * @param selector the <code>DataSelector</code> to use.
   * @param comparator the <code>Comparator</code> to use.
   * @return a <b>NEW or CLONED</b> <code>TreeSet</code> of <code>Data</code>.
   * @since jcms-5.5.0
   */
  public static <T> TreeSet<T> select(Collection<? extends T> dataColl, DataSelector selector, Comparator comparator) {
    if (dataColl == null) {
      return null;
    }
    // Special optimized case where just a clone is is needed:
    // + No selector has been given
    // + data is a TreeSet using
    // - either the default comparator if comparator is null.
    // - or the given comparator if comparator is not null.
    // @formatter:off
    if (selector == null && dataColl instanceof TreeSet &&
        ((comparator == null && ((TreeSet)dataColl).comparator() instanceof BasicStorable.CdateComparator) ||
         (comparator != null && ((TreeSet)dataColl).comparator() != null && ((TreeSet)dataColl).comparator().getClass() == comparator.getClass()))) {
      return (TreeSet) ((TreeSet) dataColl).clone();
    }
    // @formatter:on

    // Selector
    if (selector != null) {
      dataColl = new HashSet(dataColl);
      dataColl = JcmsUtil.applyDataSelector(dataColl, selector);
    }
    // Comparator
    TreeSet resultSet = new TreeSet(comparator != null ? comparator : new BasicStorable.CdateComparator());
    resultSet.addAll(dataColl);
    return resultSet;
  }

  /**
   * Apply given <code>DataSelector</code> by removing non selected <code>Data</code>
   * from the given <code>Collection</code>. <br>
   * This method <b>DOES NOT</b> create any new Collection, modify the existing one.
   * 
   * @param dataColl a <code>Collection</code> of <code>Data</code> from which to remove non selected data.
   * @param selector the <code>DataSelector</code> to apply
   * @return the <b>SAME</b> <code>Collection</code> that was given but modified.
   * @since jcms-5.5.0
   */
  public static Collection applyDataSelector(Collection dataColl, DataSelector selector) {
    if (dataColl == null || selector == null) {
      return dataColl;
    }
    for (Iterator it = dataColl.iterator(); it.hasNext();) {
      Data data = (Data) it.next();
      if (!selector.isSelected(data)) {
        it.remove();
      }
    }
    return dataColl;
  }

  // -----------------------------------------------
  // Multilanguage methods
  // -----------------------------------------------

  /**
   * Merges a multilingual (ML) map of a work copy with
   * the corresponding ML map of its main instance.
   * 
   * @param mlCopy A ML map of the work copy
   * @param mlMain A ML map of the main instance
   * @since jcms-5.5.0
   */
  public static void mergeMLMap(Map mlCopy, Map mlMain) {
    if (mlCopy == null || mlMain == null) {
      return;
    }
    for (Iterator it = mlCopy.entrySet().iterator(); it.hasNext();) {
      Map.Entry itEntry = (Map.Entry) it.next();
      Object itValue = itEntry.getValue();

      // skip empty value
      if (Util.notEmpty(itValue)) {
        continue;
      }

      // replace none empty value
      Object itLang = itEntry.getKey();
      itEntry.setValue(mlMain.get(itLang));
    }
  }

  /**
   * Returns the localized value mapped to the given language.
   * 
   * @param map the language map {lang, value}
   * @param lang the language. If null use the main language. If result null, use an arbitrary one.
   * @return the localized value mapped to the given language.
   * @since jcms-5.7.0
   */
  public static String getLangValue(Map map, String lang) {
    if (map == null || map.isEmpty()) {
      return null;
    }

    if (lang != null) {
      String label = (String) map.get(lang);
      if (Util.notEmpty(label)) {
        return label;
      }
    }

    String value = (String) map.get(channel.getLanguage());
    if (Util.notEmpty(value)) {
      return value;
    }
    // If the label is not set for the main language, we take the first label in the map
    // (we are sure there is one, because we tested map.isEmpty())
    return (String) map.values().iterator().next();
  }

  /**
   * Returns a map {lang, value} with the given localized values.
   * <br/>
   * Any HTML contained in the values is escaped (see {@link #escapeHtml(String)}).
   * 
   * @param values the values sorted according the language list.
   * @return a map {lang, value} with the given localized values.
   * @since jcms-5.7.0
   */
  public static Map<String, String> getLanguageMap(String[] values) {
    return getLanguageMap(values, true);
  }

  /**
   * Returns a map {lang, value} with the given localized values.
   * <br/>
   * The localized value HTML will be escaped if requested in parameter.
   * 
   * @param values the values sorted according the language list.
   * @param escapeHtml true to invoke escapeHtml method on each value stored in map, false to leave text as is.
   * @return a map {lang, value} with the given localized values.
   * @since jcms-7.0.2 jcms-7.1.0
   */
  public static Map<String, String> getLanguageMap(String[] values, boolean escapeHtml) {
    Map<String, String> map = new HashMap<String, String>();
    if (values == null) {
      return map;
    }

    List<String> langList = channel.getLanguageList();
    for (int i = 0; i < values.length && i < langList.size(); i++) {
      String value = escapeHtml ? escapeHtml(values[i]) : values[i];
      if (Util.notEmpty(value)) {
        map.put(langList.get(i), value);
      }
    }
    return map;
  }

  /**
   * Returns a map {lang, value} build from the given property key.
   * 
   * @param propKey the property key
   * @return a map {lang, value} build from the given property key.
   * @since jcms-5.7.0
   */
  public static Map<String, String> getLanguageMap(String propKey) {
    Map<String, String> map = new HashMap<String, String>();
    for (String lang : channel.getLanguageList()) {
      map.put(lang, glp(lang, propKey));
    }
    return map;
  }

  /**
   * Returns a list of map {lang, value} with the given localized values.
   * Example : getLanguageMapList(new String[] {"english1", "french1", "english2", "french2"}) returns
   * a list of 2 maps : m1: ({en, english1} {fr, french1}) and m2: ({en, english2} {fr, french2})
   * 
   * @param values the values sorted according the order then according the language list.
   * @return a list of map {lang, value} with the given localized values.
   * @since jcms-5.7.0
   */
  public static List getLanguageMapList(String[] values) {
    List list = new ArrayList();

    if (values == null) {
      return list;
    }

    List langList = channel.getLanguageList();
    int langListSize = langList.size();

    // If that value count is not proportional with language count, fill with null
    values = Util.trim(values);
    int mod = values.length % langListSize;
    if (mod != 0) {
      int newSize = values.length + langListSize - mod;
      String[] newValues = new String[newSize];
      System.arraycopy(values, 0, newValues, 0, values.length);
      values = newValues;
    }

    int n = values.length / langListSize;

    int k = 0;
    for (int i = 0; i < n; i++) {
      Map map = new HashMap();
      for (Iterator it = langList.iterator(); it.hasNext();) {
        String lang = (String) it.next();
        String value = values[k];
        map.put(lang, value == null ? "" : value);
        k++;
      }
      list.add(map);
    }

    return list;
  }

  /**
   * Returns a String[] which contains values of the given language map in the current site languages.
   * <p>
   * null or missing values in the map are replaced with empty string in the array.
   * 
   * @param map the language map {lang, value}
   * @return a String[] which contains ML values from given map for the current site languages, in the order
   *         specified by {@link Channel#getLanguageList()}. return null if specified map is null
   * @since jcms-5.7.0
   */
  public static String[] getLanguageArray(Map map) {

    if (Util.isEmpty(map)) {
      return null;
    }

    List langList = channel.getLanguageList();

    String[] values = new String[langList.size()];
    int i = 0;
    for (Iterator it = langList.iterator(); it.hasNext();) {
      String lang = (String) it.next();
      values[i] = Util.getString((String) map.get(lang), "");
      i++;
    }

    return values;
  }

  /**
   * This methods adds the given data to the given set if this data
   * is an instance of the given class.
   * It is mainly used by Publication.getLinkDataSet().
   * 
   * @param set the Set
   * @param clazz the class used to filter data
   * @param data the data to be added
   * @since jcms-5.0.3
   */
  public static void addData(TreeSet set, Class clazz, Data data) {

    if (set == null || clazz == null || data == null) {
      return;
    }

    if (!clazz.isAssignableFrom(data.getClass())) {
      return;
    }

    set.add(data);
  }

  /**
   * This methods adds data from an array to the given set if the data
   * is an instance of the given class.
   * It is mainly used by Publication.getLinkDataSet().
   * 
   * @param set the Set
   * @param clazz the class used to filter data
   * @param array the array of data to be added
   * @since jcms-5.0.3
   */
  public static void addData(TreeSet set, Class clazz, Data[] array) {

    if (set == null || clazz == null || array == null) {
      return;
    }

    for (int i = 0; i < array.length; i++) {
      Data itData = array[i];
      if (itData == null) {
        continue;
      }
      if (clazz.isAssignableFrom(itData.getClass())) {
        set.add(itData);
      }
    }
  }

  /**
   * This methods adds data from a collection to the given set if the data
   * is an instance of the given class.
   * It is mainly used by Publication.getLinkDataSet().
   * 
   * @param set the Set
   * @param clazz the class used to filter data
   * @param coll the collection of data to be added
   * @since jcms-5.7.0
   */
  public static void addData(TreeSet set, Class clazz, Collection coll) {

    if (set == null || clazz == null || coll == null) {
      return;
    }

    for (Iterator it = coll.iterator(); it.hasNext();) {

      Object obj = it.next();
      if (obj == null) {
        continue;
      }
      if (obj instanceof Data) {
        addData(set, clazz, (Data) obj);
      } else if (obj instanceof Data[]) {
        addData(set, clazz, (Data[]) obj);
      }
    }
  }

  /**
   * This methods adds data from a ML Map to the given set if the data
   * is an instance of the given class.
   * It is mainly used by Publication.getLinkDataSet().
   * 
   * @param set the Set
   * @param clazz the class used to filter data
   * @param mlMap the ML map of data to be added
   * @since jcms-5.7.0
   */
  public static void addData(TreeSet set, Class clazz, Map mlMap) {

    if (set == null || clazz == null || mlMap == null) {
      return;
    }

    for (Iterator it = mlMap.values().iterator(); it.hasNext();) {

      Object obj = it.next();
      if (obj == null) {
        continue;
      }
      if (obj instanceof Data) {
        addData(set, clazz, (Data) obj);
      } else if (obj instanceof Data[]) {
        addData(set, clazz, (Data[]) obj);
      }
    }
  }

  /**
   * Adds the FileDocument bound to the given filename in the given set.
   * 
   * @param set a set of FileDocument
   * @param filename the filename
   * @since jcms-6.0.1
   */
  public static void addFileDocument(Set<FileDocument> set, String filename) {

    if (set == null || filename == null) {
      return;
    }

    FileDocument doc = FileDocument.getFileDocumentFromFilename(filename);
    if (doc == null) {
      return;
    }

    set.add(doc);
  }

  /**
   * Adds the FileDocuments bound to the given filename array in the given set.
   * 
   * @param set a set of FileDocument
   * @param filenames an array of filename
   * @since jcms-6.0.1
   */
  public static void addFileDocument(Set<FileDocument> set, String... filenames) {

    if (set == null || filenames == null) {
      return;
    }

    for (String filename : filenames) {
      FileDocument doc = FileDocument.getFileDocumentFromFilename(filename);
      if (doc == null) {
        continue;
      }
      set.add(doc);
    }
  }

  /**
   * Adds the FileDocuments bound to the given filename collection in the given set.
   * 
   * @param set a set of FileDocument
   * @param filenameColl a collection of filename
   * @since jcms-7.1.1
   */
  public static void addFileDocument(Set<FileDocument> set, Collection<String> filenameColl) {

    if (set == null || filenameColl == null) {
      return;
    }

    for (String filename : filenameColl) {
      FileDocument doc = FileDocument.getFileDocumentFromFilename(filename);
      if (doc == null) {
        continue;
      }
      set.add(doc);
    }
  }

  /***
   * Adds the FileDocuments bound to the given filename and to the filename contains as value of mlMap in the given set.
   * 
   * @param set a set of FileDocument
   * @param filenameMainLanguage a filename
   * @param mlMap a map which contains filename as values
   * @since jcms-6.0.1
   */
  public static void addFileDocument(Set<FileDocument> set, String filenameMainLanguage, Map<String, String> mlMap) {

    if (set == null) {
      return;
    }

    if (filenameMainLanguage != null) {
      addFileDocument(set, filenameMainLanguage);
    }

    if (mlMap != null) {
      for (String filename : mlMap.values()) {
        FileDocument doc = FileDocument.getFileDocumentFromFilename(filename);
        if (doc == null) {
          continue;
        }
        set.add(doc);
      }
    }

  }

  /**
   * Adds the FileDocuments bound to the given array of filename and to the array of filename contains as value of mlMap in the given set.
   * 
   * @param set a set of FileDocument
   * @param filenamesMainLanguage an array of filename
   * @param mlMap a map which contains array of filename as values
   * @since jcms-6.0.1
   */
  public static void addFileDocument(Set<FileDocument> set, String[] filenamesMainLanguage, Map<String, String[]> mlMap) {

    if (set == null || mlMap == null) {
      return;
    }

    addFileDocument(set, filenamesMainLanguage);

    for (String[] array : mlMap.values()) {

      if (array == null) {
        continue;
      }
      for (String filename : array) {
        FileDocument doc = FileDocument.getFileDocumentFromFilename(filename);
        if (doc == null) {
          continue;
        }
        set.add(doc);
      }
    }
  }

  /**
   * This convenient method return the id of the given data.
   * 
   * @param data the data (may be null)
   * @return the id of the data or null if the given data is null.
   * @since jcms-6.0.0
   */
  public static String getId(Data data) {
    if (data == null) {
      return null;
    }

    return data.getId();
  }

  /**
   * Returns true if the 2 data have the same and non null ID.
   * 
   * @param d1 the first data
   * @param d2 the second data
   * @return true if the 2 data have the same and non null ID.
   * @since jcms-6.1.0
   */
  public static boolean isSameId(Data d1, Data d2) {
    if (d1 != null && d1 == d2) {
      return true;
    }
    String id1 = getId(d1);
    String id2 = getId(d2);
    if (id1 == null) {
      return false;
    }

    return id1.equals(id2);
  }

  /**
   * Compute the real language used for the given field and retrieve the
   * html language attributes to use for the this language. <br>
   * <br>
   * For example given this SmallNews:
   * <table>
   * <tr><td></td><td>English ("en")</td><td>Arabic ("ar")</td></tr>
   * <tr><td>Introduction ("intro")</td><td>an introduction</td><td></td></tr>
   * <tr><td>Content ("content")</td><td>some content</td><td>ألف</td></tr>
   * </table>
   * If you want to display the introduction in arabic, because the field is only
   * available in english, this will return the english language attribute:<br>
   * <code>getFieldLangAttributes(theSmallNews, "intro", userLang); // userLang='ar'</code><br>
   * returns <code>lang="en" xml:lang="en" dir="ltr"</code>.<br>
   * <br>
   * But if you want to display the content, it is available in the userLanguage and
   * therefore will return the arabic language attributes: <br>
   * <code>getFieldLangAttributes(theSmallNews, "content", userLang); // userLang='ar'</code><br>
   * returns <code>lang="ar" xml:lang="ar" dir="rtl"</code>.<br>
   * 
   * <b>BUT</b> because the userLang is already used to define lang attributes of
   * the html page, if the real field lang is the same, we return an empty string to
   * prevent uneeded html.
   * 
   * @param data the Publication or Data of which to check field value in given language
   * @param fieldName the name of the field to check.
   * @param userLang the language in which the field was should be displayed
   * @return a string containing the html 'lang' and 'dir' attributes without any
   *         surrounding spaces, or an empty string if advanced I18N is not enabled in JCMS.
   * @see Channel#isUsingAdvancedI18N()
   * @see Channel#getUsedLanguage(Data, String, String)
   * @see #getLangAttributes(String)
   */
  public static String getFieldLangAttributes(Data data, String fieldName, String userLang) {
    if (!channel.isUsingAdvancedI18N()) {
      return "";
    }
    String fieldLang = channel.getUsedLanguage(data, fieldName, userLang);
    // no need to specify it because the userLang has already been used as default html lang
    if (fieldLang.equals(userLang)) {
      return "";
    }
    return getLangAttributes(fieldLang);
  }

  /**
   * Retrieve the html language attributes to use for the given language.<p>
   * For example : <tt>lang="fr" xml:lang="fr" dir="ltr"</tt>
   * 
   * @param lang the language of which to generate html attributes or null to use default jcms language
   * @return a string containing the html 'lang' and 'dir' attributes without any
   *         surrounding spaces, or an empty string if advanced I18N is not enabled in JCMS.
   * @see Channel#getLanguageDirection(String)
   */
  public static String getLangAttributes(String lang) {
    if (Util.isEmpty(lang)) {
      return "";
    }
    StringBuffer sb = new StringBuffer(42); // 42 is a large maximum
    sb.append("lang=\"").append(lang).append("\" xml:lang=\"").append(lang).append("\" dir=\"");
    sb.append(channel.getLanguageDirection(lang)).append("\"");
    return sb.toString();
  }

  /**
   * Retrieve the HTML attributes to use for any HTML element surrounding feature/s which
   * are not translated in the current user language (eg : ctx menu, glp, ...) :
   * <ul>
   * <li> When the current user language is not a JCMS localization (eg: "es", "ar", ...),
   * this variable will set the proper language and directionnality (eg lang="en" dir="ltr"),</li>
   * <li> When the current user language is a valid JCMS localization (eg : "en", "fr"),
   * this variable is an empty string.the html language attributes to use for the given language.</li>
   * </ul>
   * 
   * @param lang the current language (ISO-639)
   * @return a string containing the html 'lang' and 'dir' attributes without any
   *         surrounding spaces, or an empty string if the specified language is a JCMS localized language.
   * @see Channel#getLanguageDirection(String)
   * @since jcms-7.0.0
   */
  public static String getUntranslatedZoneAttribute(String lang) {
    final String untranslatedZoneAttribute;
    if (Util.notEmpty(lang) && !channel.getJcmsLanguageList().contains(lang)) {
      untranslatedZoneAttribute = JcmsUtil.getLangAttributes(channel.getLanguage());
    } else {
      untranslatedZoneAttribute = "";
    }
    return untranslatedZoneAttribute;
  }

  /**
   * Returns the value of static variable REVISION of the given class.
   * 
   * @param clazz the class to get the revision.
   * @return the value of static variable REVISION
   * @since jcms-5.5.0
   */
  public static String getRevision(Class clazz) {
    String revision = revisionCache.getProperty(clazz.getName());
    if (revision != null) {
      return revision;
    }
    revision = "";
    try {
      Field field = clazz.getDeclaredField("REVISION");

      String cvsRevision = (String) field.get(clazz);
      revision = Util.replace(cvsRevision, "s/\\$" + "Revision" + ":\\s*(\\S+)\\s*\\$/$1/"); // Split to avoid CVS replacement!
    } catch (NoSuchFieldException ex) {
      // EMPTY
    } catch (IllegalAccessException ex) {
      // EMPTY
    } catch (Throwable ex) {
      logger.warn("An exception occured while getting revision for " + clazz, ex);
    }
    revisionCache.put(clazz.getName(), revision);
    return revision;
  }

  /**
   * Returns the value of static variable REVISION when it was packaged for this release.
   * 
   * @param clazz the class to get the revision.
   * @return the value of static variable REVISION
   * @since jcms-5.5.0
   */
  public static String getReleasedRevision(Class clazz) {
    if (releasedRevisionCache == null) {
      releasedRevisionCache = new Properties();
      InputStream input = null;
      try {
        input = new BufferedInputStream(new FileInputStream(channel.getJaliosPath("revisions.prop")));
        releasedRevisionCache.load(input);
        input.close();
      } catch (Exception ex) {
        logger.warn("Cannot load revisions.prop");
        if (input != null) {
          try {
            input.close();
          } catch (Exception ex2) {
            // EMPTY
          }
        }
      }
    }
    return releasedRevisionCache.getProperty(clazz.getName());
  }

  /**
   * Returns the list of the Jalios classes
   * 
   * @return the list of the Jalios classes.
   * @since jcms-5.5.0
   */
  public static List<Class<?>> getClassList() {
    return getClassList(channel.getWebInfPath("/lib/"), channel.getClassLoader());
  }

  /**
   * Returns the list of the Jalios classes
   * 
   * @param libDir the path for lib directory
   * @param classLoader the classLoader to be used
   * @return the list of the Jalios classes.
   * @since jcms-5.5.0
   */
  public static synchronized List<Class<?>> getClassList(String libDir, ClassLoader classLoader) {

    if (classList != null) {
      return classList;
    }

    classList = new ArrayList<Class<?>>();

    for (int i = 0; i < jarNames.length; i++) {
      File jarFile = new File(libDir, jarNames[i]);
      if (!jarFile.exists()) {
        logger.warn("jar file WEB-INF/lib/" + jarNames[i] + " could not be found ");
        continue;
      }

      ZipInputStream zis = null;
      try {
        FileInputStream fis = new FileInputStream(jarFile);
        zis = new ZipInputStream(new BufferedInputStream(fis));
        ZipEntry entry = null;

        while ((entry = zis.getNextEntry()) != null) {

          // Skip directory.
          if (entry.isDirectory()) {
            continue;
          }

          // Get the class name
          String className = entry.getName();
          if (className.indexOf('$') >= 0) {
            continue;
          }
          // IMPORTANT : Make sure we skip the following classes as it would start the Channel !!!
          // @formatter:off
          if (className.indexOf("JcmsTestCase") >= 0 || className.indexOf("JcmsTestCase4") >= 0 ||
              className.indexOf("JcmsTestCaseBootstrapper") >= 0 || className.indexOf("JcmsTestCaseUtil") >= 0){
            continue;
          }
          // @formatter:on
          className = Util.getClassNameFromFilename(className);

          // Load the class
          Class<?> clazz = null;
          try {
            // CAUTION:OD: Use forName(..., classLoader) instead of classLoader.loadClass(className)
            // to be compliant with JDK 1.6. In JDK 1.6, classLoader.loadClass() no more support
            // array class such as "[Ljava.lang.String" (throw ClassNotFoundException).
            // For details, see Java Bug #6434149:
            // http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=6434149
            // and JCMS discussion in JX forum:
            // http://support.jalios.com/jcms/jx_31762/zone-de-texte-riche-bug
            clazz = Class.forName(className, true, classLoader);
          } catch (Throwable th) {
            // EMPTY
          }
          if (clazz == null) {
            continue;
          }

          classList.add(clazz);
        } // end of while
      } catch (Exception ex) {
        logger.warn("Exception occured while loading \"" + jarFile + "\"", ex);
      } finally {
        if (zis != null) {
          try {
            zis.close();
          } catch (IOException ex) {
          }
        }
      }
    }

    return classList;
  }

  /**
   * Generates an XHtml meta tag (i.e. <meta name='...' content='...' />
   * 
   * @param name the name attribute
   * @param content the content attribute
   * @return an XHtml meta tag.
   * @since jcms-5.5.0
   */
  public static String getXHtmlMetaTag(String name, String content) {
    return getXHtmlMetaTag("name", name, content);
  }

  /**
   * Generates an XHtml meta http-equiv tag (i.e. <meta http-equiv='...' content='...' />
   * 
   * @param httpEquiv the httpEquiv attribute
   * @param content the content attribute
   * @return an XHtml http-equiv meta tag.
   * @since jcms-5.7.0
   */
  public static String getXHtmlMetaHttpEquivTag(String httpEquiv, String content) {
    return getXHtmlMetaTag("http-equiv", httpEquiv, content);
  }

  /**
   * Generates a generic XHtml meta tag (key name is configurable).
   * 
   * @param keyName the name of the key (eg "name", or "http-equiv")
   * @param key the key attribute
   * @param content the content attribute
   * @return an XHtml meta tag.
   * @since jcms-5.7.0
   */
  public static String getXHtmlMetaTag(String keyName, String key, String content) {

    if (Util.isEmpty(key)) {
      return "";
    }

    if (content == null) {
      content = "";
    }

    return "<meta " + keyName + "=\"" + key + "\" content=\"" + XmlUtil.normalize(content) + "\" />";
  }

  /**
   * Returns some statistics about a given publication type.
   * The returns array contains
   * <li>[0]: the publication count
   * <li>[1]: in-progress publication count
   * <li>[2]: scheduled publication count
   * <li>[3]: published publication count
   * <li>[4]: expired publication count
   * <li>[5]: restricted publication count
   * <li>[6]: audienced publication count
   * 
   * @param clazz the class
   * @param mbr the member who performs the request
   * @param ws the workspace
   * @return an int[] which contains the informations.
   * @since jcms-5.5.0
   */
  public static int[] getPublicationDetails(Class clazz, Member mbr, Workspace ws) {
    Set pubSet = channel.getPublicationSet(clazz, mbr, false, ws);
    Set docSet = channel.getPublicationSet(FileDocument.class, mbr, false, ws);
    Set pubWithoutDocSet = Util.subSet(pubSet, docSet);
    DataSelector inProgressSelector = Publication.getPstatusSelector(WorkflowManager.SCHEDULED_PSTATUS - 1, -1);
    DataSelector scheduledSelector = Publication.getPstatusSelector(WorkflowManager.SCHEDULED_PSTATUS);
    DataSelector publishedSelector = Publication.getPstatusSelector(WorkflowManager.PUBLISHED_PSTATUS);
    DataSelector expiredSelector = Publication.getPstatusSelector(WorkflowManager.EXPIRED_PSTATUS);
    DataSelector readRightSelector = Publication.getReadRightSelector();
    DataSelector audienceRightSelector = Publication.getAudienceRightSelector();
    DataSelector hiddenDocSelector = Publication.getPstatusSelector(FileDocument.HIDDEN_PSTATUS);

    int[] res = new int[8];
    res[0] = pubSet.size() + HibernateUtil.queryCount(clazz, "workspaceId", ws.getId());
    res[1] = select(pubWithoutDocSet, inProgressSelector, null).size();
    res[2] = select(pubSet, scheduledSelector, null).size();
    res[3] = select(pubSet, publishedSelector, null).size();
    res[4] = select(pubSet, expiredSelector, null).size();
    res[5] = select(pubSet, readRightSelector, null).size();
    if (AudienceRights.getInstance().isEnabled()) {
      res[6] = select(pubSet, audienceRightSelector, null).size();
    }
    res[7] = select(docSet, hiddenDocSelector, null).size();
    return res;
  }

  static Pattern externalUrlPattern;
  static {
    Perl5Compiler regexpCompiler = new Perl5Compiler();
    try {
      externalUrlPattern = regexpCompiler.compile("(((https?)|(ftp))://[^\\s'\"\\](),<]+)");
    } catch (org.apache.oro.text.regex.MalformedPatternException ex) {
      logger.warn("An exception occured while compiling urlPattern pattern", ex);
    }

  }

  /**
   * Returns all the external URL contained in the given text.
   * 
   * @param text the text to be parsed.
   * @return all the external URL contained in the given text.
   * @since jcms-5.5.0
   */
  public static Set extractExternalUrl(String text) {
    Set set = new HashSet();

    if (Util.isEmpty(text)) {
      return set;
    }

    String prefixUrl = channel.getUrl();

    Perl5Matcher matcher = new Perl5Matcher();
    matcher.setMultiline(true);
    PatternMatcherInput input = new PatternMatcherInput(text);
    while (matcher.contains(input, externalUrlPattern)) {
      MatchResult results = matcher.getMatch();
      String url = results.group(0);
      if (url.startsWith(prefixUrl)) {
        continue;
      }
      set.add(url);
    }

    return set;
  }

  // ------------------------------------------------------------
  // Status page access checks
  // ------------------------------------------------------------

  private static String[] ipsAuthorizedToAccessStatus = null;

  /**
   * Initialize the internal array of string containg all IPs addresses allowed
   * to access status (through status.jsp - statusXml.jsp)
   * 
   * @param context the required
   */
  private static void initIPsAuthorizedToAccessStatus(ServletContext context) {
    if (context == null) {
      throw new IllegalArgumentException("context argument cannot be null");
    }
    Properties props = new Properties();

    // Load properties
    FileInputStream fis = null;
    try {
      File statusPropFile = new File(context.getRealPath(JcmsConstants.STATUS_PROP));
      if (statusPropFile.exists()) {
        fis = new FileInputStream(statusPropFile);
        props.load(fis);
        ipsAuthorizedToAccessStatus = Util.split(props.getProperty("authorized-ips"), " ");
      } else {
        ipsAuthorizedToAccessStatus = new String[] { "127.0.0.1" };
      }
    } catch (Exception ex) {
      logger.warn("Could not open property file status.prop", ex);
    } finally {
      if (fis != null) {
        try {
          fis.close();
        } catch (java.io.IOException ex2) {
          /* EMPTY */ }
      }
    }
  }

  /**
   * Check if the given request is allowed to access jcms status page,
   * (admin/admin.jsp, admin/statusXml.jsp).<br>
   * <br>
   * This method can be called even if JCMS has not been started,
   * as long as a valid ServletContext is given.<br>
   * Valid IPs addresses are read from the file "WEB-INF/data/status.prop" the first
   * time this method is called.
   * 
   * @param context a required ServletContext (used to retrieve file path of the
   *          status.prop file inside the webapp).
   * @param request the request of which to check access to status, must not be null.
   * @return true if the request is allowed, false otherwise.
   */
  public static boolean isAuthorizedToAccessStatusInfo(ServletContext context, HttpServletRequest request) {
    if (request == null) {
      throw new IllegalArgumentException("request argument cannot be null");
    }

    // Check if the logged member is admin (what ever the IP is)
    Member loggedMember = (Member) request.getAttribute("loggedMember");
    if (loggedMember != null && loggedMember.isAdmin()) {
      return true;
    }

    if (ipsAuthorizedToAccessStatus == null) {
      initIPsAuthorizedToAccessStatus(context);
      if (ipsAuthorizedToAccessStatus == null) {
        return false;
      }
    }
    String ip = request.getRemoteAddr();
    boolean found = false;
    for (int i = 0; i < ipsAuthorizedToAccessStatus.length; i++) {
      if (ip.startsWith(ipsAuthorizedToAccessStatus[i])) {
        found = true;
        break;
      }
    }
    return found;
  }

  /**
   * Returns the webhelp URI of the user's guide home page.
   * 
   * @param request the HttpRequest
   * @return the webhelp URI of the user's guide home page.
   * @since jcms-5.6.1
   */
  public static String getWebHelpHome(HttpServletRequest request) {
    return getWebHelpUriAndIcon(request, null, true)[0];
  }

  /**
   * Returns the webhelp URI of the given entry.
   * 
   * @param request the HttpRequest
   * @param entry the entry to display the documentation
   * @return the webhelp URI of the user's guide of the given entry
   * @since jcms-5.6.1
   */
  public static String getWebHelpUri(HttpServletRequest request, String entry) {
    return getWebHelpUriAndIcon(request, entry, false)[0];
  }

  /**
   * Returns a String[] which contains the webhelp URI and the webhelp icon.
   * 
   * @param request the HttpRequest (to determine the JSP)
   * @param entry the WebHelp entry to access (if empty, find the entry from the request)
   * @param home if true display user's guide home page
   * @return a String[] which contains the webhelp URI and the webhelp icon.
   * @since jcms-5.6.1
   */
  public static String[] getWebHelpUriAndIcon(HttpServletRequest request, String entry, boolean home) {

    // Get the lang
    String userLang = (String) request.getAttribute("userLang");

    // Get the WebHelp entry (if not provided)
    if (entry == null) {
      entry = (String) request.getAttribute("whEntry");
      if (Util.isEmpty(entry)) {
        entry = ServletUtil.getJsp(request);
      }

      String suffix = (String) request.getAttribute("whSuffix");
      if (Util.notEmpty(suffix)) {
        entry += "." + suffix;
      }
    }

    // Get WebHelp URI
    String whDir = channel.getProperty("wh.dir", "/jalios/webhelp/") + userLang + "/";
    String iconUri = ICON_WH_BOOK_OPEN;
    String whUri = channel.getProperty("wh." + entry, null);
    if (home || whUri == null) {
      whUri = "jcms.htm";
      iconUri = ICON_WH_BOOK_CLOSED;
    }
    whUri = ServletUtil.getContextPath(request) + whDir + whUri;

    return new String[] { whUri, iconUri };
  }
  // ------------------------------------------------------------
  // URLs generation (Standard / Descriptive)
  // ------------------------------------------------------------

  /**
   * Return the display url of JCMS.
   * For example, "jcms/" if descriptive URLs are enabled, "display.jsp" otherwise).
   * 
   * @return a path relative to the webapp path, (no '/' at begining).
   * @since jcms-5.7.0
   */
  public static String getDisplayUrl() {
    final String url;
    if (DescriptiveURLs.isEnabled()) {
      url = DescriptiveURLs.getServletPath();
    } else {
      url = "display.jsp";
    }
    return url;
  }

  /**
   * Return the URL to access the given data in the front office. <br>
   * If descriptive URLs are enabled in JCMS ({@link DescriptiveURLs})
   * the url will be generated using {@link DescriptiveURLs#getDescriptiveURL(Data, Locale)}
   * otherwise, a standard "display.jsp?id=" URL is generated. <br>
   * <br>
   * If the given data is null, the returned url will be direct access to servlet path
   * (for example, "jcms/" if descriptive URLs are enabled, "display.jsp" otherwise),
   * see {@link #getDisplayUrl()}.<br>
   * <br>
   * Warning: The return url may or may not contains a query string
   * 
   * @param data the data for which the URL will be generated
   * @param locale the {@link Locale} to use for I18N and localized operations
   * @return a path relative to the webapp path, (no '/' at begining).
   * @since jcms-5.7.0
   * @see Data#getDisplayUrl(Locale)
   */
  public static String getDisplayUrl(Data data, Locale locale) {
    if (data == null) {
      return getDisplayUrl();
    }

    return data.getDisplayUrl(locale);
  }

  /**
   * Return the URL to access the given data in the front office with the added given param string. <br>
   * If descriptive URLs are enabled in JCMS ({@link DescriptiveURLs})
   * the url will be generated using {@link DescriptiveURLs#getDescriptiveURL(Data, Locale)}
   * otherwise, a standard "display.jsp?id=" URL is generated. <br>
   * <br>
   * Examples :
   * 
   * <pre>
   * String link = JcmsUtil.getDisplayUrl(myFileDoc, userLocale, "details=true&amp;a=b", true);
   * </pre>
   * 
   * <pre>
   * String redirect = JcmsUtil.getDisplayUrl(homeCat, locale, "mid=" + mbr.getId() + "&jsp=" + ResourceHelper.getMemberDisplay(), false);
   * </pre>
   * 
   * @param data the data for which the URL will be generated
   * @param locale the {@link Locale} to use for I18N and localized operations
   * @param params the parameter string to add, it should not start
   *          with any ampersand ('&' or '&amp;amp;') as it will be added automatically
   * @param escapeAmpersand will use "&amp;amp; true, "&" if false
   * @return a path relative to the webapp path, (no '/' at begining).
   * @since jcms-5.7.0
   */
  public static String getDisplayUrl(Data data, Locale locale, String params, boolean escapeAmpersand) {
    String displayUrl = getDisplayUrl(data, locale);
    if (Util.isEmpty(params)) {
      return displayUrl;
    }

    StringBuffer sb = new StringBuffer(displayUrl.length() + 8 + params.length());
    sb.append(displayUrl);
    if (displayUrl.indexOf('?') == -1) {
      sb.append("?");
    } else {
      sb.append(escapeAmpersand ? "&amp;" : "&");
    }
    sb.append(params);

    return sb.toString();
  }

  /**
   * Return the URL to access the given data in the front office, while
   * updating/adding or removing a set of parameters. <br>
   * If descriptive URLs are enabled in JCMS ({@link DescriptiveURLs})
   * the url will be generated using {@link DescriptiveURLs#getDescriptiveURL(Data, Locale)}
   * otherwise, a standard "display.jsp?id=" URL is generated. <br>
   * 
   * @param data the data for which the URL will be generated
   * @param locale the {@link Locale} to use for I18N and localized operations
   * @param request the current {@link HttpServletRequest} of which to
   *          update/add/remove params.
   *          If null, this method as the same behavior as {@link #getDisplayUrl(Data, Locale)}.
   * @param names an array of String which contains the name of the parameters
   * @param newValues an array of String which contains the values of
   *          the parameters (if newValue[i] is null the names[i] parameter is
   *          removed)
   * @param removePatterns any parameter name containing one of the given Strings will be removed from the parameters' list
   * @param escapeAmpersand will use "&amp;amp; true, "&" if false
   * @return a path relative to the webapp path, (no '/' at begining).
   * @since jcms-5.7.0
   */
  public static String getDisplayUrlWithUpdatedParams(Data data, Locale locale, HttpServletRequest request, String[] names, String[] newValues,
      String[] removePatterns, boolean escapeAmpersand) {

    StringBuffer sb = new StringBuffer(ServletUtil.URL_INIT_SIZE);

    // Display URL
    sb.append(getDisplayUrl(data, locale));

    // QueryString
    if (request != null && (names != null || newValues != null || removePatterns != null)) {
      DescriptiveURLs.appendUpdatedParamsNoID(sb, ServletUtil.getOrderedParameterMap(request), names, newValues, removePatterns, escapeAmpersand);
    }

    return sb.toString();
  }

  /**
   * Return the absolute URL to access the given data in the front office, while
   * updating/adding or removing a set of parameters. <br>
   * If descriptive URLs are enabled in JCMS ({@link DescriptiveURLs})
   * the url will be generated using {@link DescriptiveURLs#getDescriptiveURL(Data, Locale)}
   * otherwise, a standard "display.jsp?id=" URL is generated. <br>
   * 
   * @param data the data for which the URL will be generated
   * @param locale the {@link Locale} to use for I18N and localized operations
   * @param request the current {@link HttpServletRequest} of which to
   *          update/add/remove params.
   *          If null, this method as the same behavior as {@link #getDisplayUrl(Data, Locale)}.
   * @param names an array of String which contains the name of the parameters
   * @param newValues an array of String which contains the values of
   *          the parameters (if newValue[i] is null the names[i] parameter is
   *          removed)
   * @param removePatterns any parameter name containing one of the given Strings will be removed from the parameters' list
   * @param escapeAmpersand will use "&amp;amp; true, "&" if false
   * @return a full and absolute URL
   * @since jcms-5.7.0
   */
  public static String getAbsDisplayUrlWithUpdatedParams(Data data, Locale locale, HttpServletRequest request, String[] names, String[] newValues,
      String[] removePatterns, boolean escapeAmpersand) {

    StringBuffer sb = new StringBuffer(ServletUtil.URL_INIT_SIZE);

    // Base URL
    sb.append(ServletUtil.getBaseUrl(request));

    // Display URL
    sb.append(getDisplayUrl(data, locale));

    // QueryString
    if (request != null && (names != null || newValues != null || removePatterns != null)) {
      DescriptiveURLs.appendUpdatedParamsNoID(sb, ServletUtil.getOrderedParameterMap(request), names, newValues, removePatterns, escapeAmpersand);
    }

    return sb.toString();
  }

  /**
   * Return the URL to access the given data in the front office, while
   * adding the given parameter. <br>
   * If descriptive URLs are enabled in JCMS ({@link DescriptiveURLs})
   * the url will be generated using {@link DescriptiveURLs#getDescriptiveURL(Data, Locale)}
   * otherwise, a standard "display.jsp?id=" URL is generated. <br>
   * 
   * @param data the data for which the URL will be generated
   * @param locale the {@link Locale} to use for I18N and localized operations
   * @param request the current {@link HttpServletRequest} of which to
   *          update/add/remove params.
   *          If null, this method as the same behavior as {@link #getDisplayUrl(Data, Locale)}.
   * @param name a String which contains the name of the parameter to add
   * @param value a String which contains the value of the parameter to add
   *          (if value is null the parameter is not added, but it wont be removed from existing params)
   * @param escapeAmpersand will use "&amp;amp; true, "&" if false
   * @return a path relative to the webapp path, (no '/' at begining).
   * @since jcms-5.7.0
   */
  public static String getDisplayUrlWithAddedParam(Data data, Locale locale, HttpServletRequest request, String name, String value,
      boolean escapeAmpersand) {

    StringBuffer sb = new StringBuffer(ServletUtil.URL_INIT_SIZE);

    // Display URL
    sb.append(getDisplayUrl(data, locale));

    // QueryString
    if (request != null) {
      DescriptiveURLs.appendUpdatedParamsNoID(sb, ServletUtil.getOrderedParameterMap(request), null, null, null, escapeAmpersand);
    }

    // Added params (excepting if its id, in which case it had to be manage trough data)
    if (name != null && value != null && !name.equals("id")) {
      if (sb.indexOf("?") == -1) {
        sb.append("?");
      } else {
        sb.append(escapeAmpersand ? "&amp;" : "&");
      }
      sb.append(ServletUtil.encodeURL(name));
      sb.append('=');
      sb.append(ServletUtil.encodeURL(value));
    }

    return sb.toString();
  }

  /**
   * Retrieve I18N strings from XML and properties.
   * <p>
   * Build a map containing ISO-639 language code as the key, and I18N strings as the value :
   * <ul>
   * <li>it first reads the content of the XML children element matching the specified xml tag name (which must have xml:lang attribute),</li>
   * <li>then for each *localized* languages of the site, it read the specified language property it available.</li>
   * </ul>
   * For example, given the following XML Element and properties :
   * 
   * <pre>
   *   &lt;foobar&gt;
   *    &lt;label xml:lang="en"&gt; A foo bar&lt;/label&gt;
   *    &lt;label xml:lang="fr"&gt; Un foo bar&lt;/label&gt;
   *   &lt;/foobar&gt;
   *   
   * [es.prop] whatever.foobar.label: Uno foo bar
   * [de.prop] whatever.foobar.label: Ein foo bar
   * </pre>
   * 
   * This invocation :
   * 
   * <pre>
   *   JcmsUtil.getLangMap(foobarXmlElement, "label", "whatever.foobar.label", false null);
   * </pre>
   * 
   * will result in the following map (if the site has been configure in en/fr/es/de/it languages)
   * 
   * <pre>
   * {
   *   "en" -&gt; "A foo bar",
   *   "fr" -&gt; "Un foo bar",
   *   "es" -&gt; "Uno foo bar",
   *   "de" -&gt; "Ein foo bar"
   * }
   * </pre>
   * 
   * If <code>forceMainLanguageValue</code> is set to <code>false</code>, this method is the same as :
   * 
   * <pre>
   * Map<String, String> langMap = XmlUtil.getLangChildrenMap(xmlElt, xmlChildName);
   * langMap = JcmsUtil.getLangMapUpdatedFromProperties(langMap, langProperty);
   * </pre>
   * 
   * Setting <code>forceMainLanguageValue</code> argument to <code>true</code> will
   * ensure that a non null value is defined for the main language of the site,
   * either by using another arbitrary language value if possible, or else by
   * setting string value "???". <br/>
   * In such case <code>missingMainValueInfo</code> is used as a log error message prefix.
   * 
   * @param xmlElt the JDOM XML Element in which to look for I18N strings
   * @param xmlChildrenName the name of the XML children tags to read
   * @param langProperty the name of the property to read
   * @param forceMainLanguageValue set to true to ensure that a value has been defined for the main language
   * @param missingMainValueInfo information message to include in log if value was missing in main language (used only when forceMainLanguageValue is
   *          true)
   * @return a new HashMap with an ISO-639 language code as the key, and I18N strings as the value.
   * @see XmlUtil#getLangChildrenMap(Element, String)
   * @see #getLangMapUpdatedFromProperties(Map, String)
   * @since jcms-8.0.0
   */
  public static Map<String, String> getLangMap(Element xmlElt, String xmlChildrenName, String langProperty, boolean forceMainLanguageValue,
      String missingMainValueInfo) {
    Map<String, String> langMap = XmlUtil.getLangChildrenMap(xmlElt, xmlChildrenName);
    langMap = JcmsUtil.getLangMapUpdatedFromProperties(langMap, langProperty);

    if (forceMainLanguageValue) {
      // Check that mainLanguage entry is not empty
      String mainLang = channel.getLanguage();
      if (langMap == null || langMap.get(mainLang) != null) {
        return langMap;
      }

      // @formatter:off
      logger.error(missingMainValueInfo + ": I18N strings in main language ("+mainLang+") could not be found, " +
      		"neither in XML children element '" + xmlChildrenName + "' of '" + (xmlElt != null ? xmlElt.getName() : null) + "', " +
      	  "nor in language property " + langProperty + ". Another arbitrary language value will be used if available, or default string '???' will be set.");
      // @formatter:on

      // if no label exists for main language, a label is taken with arbitrary language
      String label = "???";
      if (!langMap.isEmpty()) {
        label = (langMap.entrySet().iterator().next()).getValue();
      }
      langMap.put(mainLang, label);
    }

    return langMap;
  }

  /**
   * Returns the map of the children of <code>elt</code> with their <code>xml:lang</code> as key.
   * If the entry for main language does not exists, an error message is append in the logs and
   * the value "???" is assigned for the main language
   * 
   * @param elt the Element
   * @param childName the name of the child
   * @param info the string to prepend in the error log message
   * @return a map of {<code>lang</code>, <code>elt</code>}.
   * @see XmlUtil#getLangChildrenMap(Element, String)
   * @since jcms-5.7.0
   */
  public static Map<String, String> getLangChildrenMap(Element elt, String childName, String info) {
    Map<String, String> map = XmlUtil.getLangChildrenMap(elt, childName);

    // Check that mainLanguage entry is not empty
    String mainLang = channel.getLanguage();
    if (map == null || map.get(mainLang) != null) {
      return map;
    }

    logger.error(info + ": '" + childName + "' element does not contain entry for the main language (" + mainLang + ").");

    // if no label exists for main language, a label is taken with arbitrary language
    String label = "???";
    if (!map.isEmpty()) {
      label = (map.entrySet().iterator().next()).getValue();
    }
    map.put(mainLang, label);
    return map;
  }

  /**
   * Update the specified map of I18N string (lang to string) with values
   * read from the specified property (if available), for all localized languages of the site.
   * 
   * I18N strings in the specified map are updated only if they did not exist.
   * 
   * @param langMap the map to update (if null a new one is created and returned)
   * @param property the property to read
   * @return the update lang map or a new one if it was specified null
   * @since jcms-8.0.0
   */
  public static Map<String, String> getLangMapUpdatedFromProperties(Map<String, String> langMap, String property) {
    langMap = langMap != null ? langMap : new HashMap<String, String>();

    if (channel == null) {
      return langMap;
    }
    for (String lang : channel.getLocalizedLanguageList()) {
      if (langMap.containsKey(lang)) {
        continue;
      }

      String labelFromProperty = JcmsUtil.glpNoDefault(lang, property);
      if (labelFromProperty == null) {
        continue;
      }

      langMap.put(lang, labelFromProperty);
    }

    return langMap;
  }

  /**
   * Returns the relative path of directory to use for the new document to be created.
   * 
   * @param baseDir the base directory (eg /usr/local/tomcat/webapps/jcms/)
   * @param subDir an optionnal sub directory, may be null (eg upload/docs/image/gif/)
   * @return String the directory path, relative to the webapp, with the trailing slash ("/").
   *         e.g: upload/docs/image/gif/2006/11/
   * @since jcms-5.7.0
   * @see IOUtil#getNewDirectory(File, String, long)
   */
  public static String getNewDirectory(File baseDir, String subDir) {
    long maxFilesPerDir = Util.toLong(channel.getProperty("channel.max-files-per-dir"), 4096); // TODO OD could be cached
    return IOUtil.getNewDirectory(baseDir, subDir, maxFilesPerDir);
  }

  /*
   * Set in the session the good message explaining the data write is disabled.
   *  
   * @param request The current Http Request
   * @param userLang The lang of the current user
   */
  public static void setDataWriteDisabledMsg(HttpServletRequest request, String userLang) {
    // @formatter:off
    JcmsContext.setWarningMsgSession(glp(userLang,
                                         "ui.adm.data-write.full-msg", 
                                         Util.getString(channel.getProperty("channel.data-write.status"),
                                                     glp(userLang, "ui.adm.data-write.default-msg"))),
                            request);
    // @formatter:on
  }

  /**
   * Check is data writes are disabled.
   * If data writes are disabled, set in session a warning msg and returns true.
   * If data writes are disabled and redirect not empty, a redirect is done to this url.
   * 
   * @param request The current request
   * @param response The current response
   * @param redirect The wished redirect
   * @param userLang The lang of the current user
   * @return true is data writes are enabled
   * @throws IOException If an error occurs
   */
  public static boolean checkDataWriteDisabled(HttpServletRequest request, HttpServletResponse response, String redirect, String userLang)
      throws IOException {
    if (channel.isDataWriteEnabled()) {
      return false;

    }
    JcmsUtil.setDataWriteDisabledMsg(request, userLang);
    if (Util.notEmpty(redirect)) {
      JcmsContext.sendRedirect(redirect, request, response);
    }
    return true;
  }

  // -----------------------------------------------------------------
  // Generic XML serialization/deserialization of I18N fields for DB
  // -----------------------------------------------------------------

  private static final String MULTILINGUAL_XML_ROOTELT_NAME = "ml";
  private static final String MULTILINGUAL_XML_VALUEELT_NAME = "value";

  /**
   * Convert a multilingual field to an xml representation.
   * <p>
   * Example :
   * 
   * <pre>
   * String xml = convertMultilingualFieldToXml(pub.getTitle(), pub.getTitleML());
   * </pre>
   * 
   * Will lead to the following xml :
   * 
   * <pre>
   *  &lt;?xml version='1.0' encoding='UTF-8'?>
   *  &lt;ml>
   *   &lt;value xml:lang="en">Title&lt;/value>
   *   &lt;value xml:lang="fr">Titre&lt;/value>
   *  &lt;/ml>
   * </pre>
   * 
   * @param mainValue the value of the field in the channel main language.
   * @param valuesMap the values of the field in other languages (ISO-639 language code as the key, value of field in value)
   * @return an xml string.
   * @since jcms-6.0
   */
  public static String convertMultilingualFieldToXml(final String mainValue, final Map<String, String> valuesMap) {
    Element rootElt = convertMultilingualFieldToXmlElement(mainValue, valuesMap);
    String xml = XmlUtil.XML_HEADER + XmlUtil.getXmlString(rootElt);
    return xml;
  }

  /**
   * Convert a multilingual field to an xml representation.
   * <p>
   * Example :
   * 
   * <pre>
   * String xml = convertMultilingualFieldToXml(pub.getTitle(), pub.getTitleML());
   * </pre>
   * 
   * Will lead to the following xml :
   * 
   * <pre>
   *  &lt;?xml version='1.0' encoding='UTF-8'?>
   *  &lt;ml>
   *   &lt;value xml:lang="en">Title&lt;/value>
   *   &lt;value xml:lang="fr">Titre&lt;/value>
   *  &lt;/ml>
   * </pre>
   * 
   * @param mainValue the value of the field in the channel main language.
   * @param valuesMap the values of the field in other languages (ISO-639 language code as the key, value of field in value)
   * @return a JDOM xml Element.
   * @since jcms-6.0
   */
  public static Element convertMultilingualFieldToXmlElement(final String mainValue, final Map<String, String> valuesMap) {
    Element tmpElt = null;
    Element rootElt = new Element(MULTILINGUAL_XML_ROOTELT_NAME);

    // main language
    tmpElt = getMultilingualXmlElement(channel.getLanguage(), mainValue);
    if (tmpElt != null) {
      rootElt.addContent(tmpElt);
    }

    // other languages
    if (Util.notEmpty(valuesMap)) {
      for (Map.Entry<String, String> entry : valuesMap.entrySet()) {
        String curLang = entry.getKey();
        String curValue = entry.getValue();
        tmpElt = getMultilingualXmlElement(curLang, curValue);
        if (tmpElt != null) {
          rootElt.addContent(tmpElt);
        }
      }
    }

    return rootElt;
  }

  /**
   * Create an JDOM XML Element from the specified lang and value.
   * 
   * <pre>
   *   &lt;value xml:lang="en">Title&lt;/value>
   * </pre>
   * 
   * @param lang the language of the value (required)
   * @param value the value of the field (required)
   * @return an Element or null if the either the lang or value is empty.
   * @since jcms-6.0
   */
  private static Element getMultilingualXmlElement(final String lang, final String value) {
    if (Util.isEmpty(lang) || Util.isEmpty(value)) {
      return null;
    }
    Element elt = new Element(MULTILINGUAL_XML_VALUEELT_NAME);
    elt.setAttribute("lang", lang, Namespace.XML_NAMESPACE);
    elt.setText(value);
    return elt;
  }

  /**
   * Converts an xml string to an hashmap of values.
   * <p>
   * Example given the following text :
   * 
   * <pre>
   *  &lt;?xml version='1.0' encoding='UTF-8'?>
   *  &lt;ml>
   *   &lt;value xml:lang="en">Title&lt;/value>
   *   &lt;value xml:lang="es">Titulo&lt;/value>
   *   &lt;value xml:lang="fr">Titre&lt;/value>
   *  &lt;/ml>
   * </pre>
   * 
   * Returns a map containing all values :
   * 
   * <pre>
   * {
   *   "en" => "Title",
   *   "fr" => "Titre",
   *   "es" => "Titulo"
   * }
   * </pre>
   * 
   * @param xml an xml representation of a multilingual field, as returned by {@link #convertMultilingualFieldToXml(String, Map) }
   * @return a map withy language code as the key and field value as the value. Return null if no field value could be found or if the xml could not
   *         be parsed.
   * @since jcms-6.0
   */
  public static Map<String, String> convertMultilingualFieldXmlToMap(final String xml) {
    try {
      Document doc = XmlUtil.getDocument(xml);
      return XmlUtil.getLangChildrenMap(doc.getRootElement(), MULTILINGUAL_XML_VALUEELT_NAME);
    } catch (Exception ex) {
    }
    return null;
  }

  /**
   * Check if the given JDBC connection can be established.
   * 
   * @param driver the class name of the JDBC driver
   * @param url a database url of the form jdbc:subprotocol:subname
   * @param user the database user on whose behalf the connection is being made
   * @param password the user's password
   * @throws Exception exception
   * @since jcms-6.0.0
   */
  public static void checkJDBCConnection(String driver, String url, String user, String password) throws Exception {

    channel.getClass(driver);
    Connection connection = DriverManager.getConnection(url, user, password);
    if (connection != null) {
      connection.close();
    }
  }

  /**
   * Returns true if the given urid is valid.
   * 
   * @param urid the urid
   * @return true if the given urid is valid.
   * @since jcms-6.0.0
   */
  public static boolean checkUrid(String urid) {
    if (!StoreUtil.checkUrid(urid)) {
      return false;
    }

    if (DBConstants.DBDATA_URID.equals(urid + "_")) {
      return false;
    }

    return true;
  }

  /**
   * Retrieves the virtual id corresponding to the specified data id.
   * <p>
   * 
   * Example :<br>
   * <code>JcmsUtil.getVirtualId("$jcmsplugin.myplugin.my-virtual-id", "c_42");</code> will returns :
   * <ul>
   * <li><code>"$jcmsplugin.myplugin.my-virtual-id"</code> if this virtual could be resolved to id <code>"c_42"</code></li>
   * <li><code>"c_42"</code> otherwise</li>
   * </ul>
   * 
   * @param virtualId a virtual id such as "$jcmsplugin.myplugin.my-virtual-id" (if anything else than a virtual id is specified, the realId is
   *          returned)
   * @param id a data id
   * @return the virtual id if it matches the specified id, otherwise, the id.
   * @since jcms-5.7.5
   */
  public static String getVirtualId(String virtualId, String id) {
    if (Util.notEmpty(virtualId) && virtualId.charAt(0) == '$') {
      String correspondingId = channel.getProperty(virtualId);
      if (correspondingId != null && correspondingId.equals(id)) {
        return virtualId;
      }
    }
    return id;
  }

  /**
   * Convenient method to keep virtual id when set.
   * It allows to keep virtual Ids in a list (actually a string of id seperated by '|') if
   * these ids has not been modified in a form.
   * 
   * A new String is constructed form the two in parameters.
   * 
   * If the number of items in the list is not the same, the second parameter is returned.
   * 
   * If the first parameter is a virtual ID, corresponding to the id in second parameter,
   * the first parameter is returned.
   * 
   * @param stringWithPossiblyVirtualIds it is a value already used, possibly a virtual id
   * @param stringOfRealIds it is the value newly set. updated or not.
   * @return one of the two parameter
   * @since jcms-5.7.5
   */
  public static String getVirtualIdList(String stringWithPossiblyVirtualIds, String stringOfRealIds) {
    if (Util.isEmpty(stringWithPossiblyVirtualIds) || Util.isEmpty(stringOfRealIds)) {
      return stringOfRealIds;
    }
    String[] arrayOfPossiblyVirtualId = stringWithPossiblyVirtualIds.split("\\|");
    String[] arrayOfRealId = stringOfRealIds.split("\\|");

    if (arrayOfPossiblyVirtualId.length != arrayOfRealId.length) {
      return stringOfRealIds;
    }

    StringBuffer buffer = new StringBuffer();
    for (int i = 0; i < arrayOfRealId.length; i++) {
      if (i != 0) {
        buffer.append('|');
      }
      buffer.append(getVirtualId(arrayOfPossiblyVirtualId[i], arrayOfRealId[i]));
    }
    return buffer.toString();
  }

  /**
   * Returns the count of members having the worker ability.
   * 
   * @return the count of members having the worker ability.
   * @since jcms-6.0.0
   */
  public static int getWorkerCount() {
    Set workerSet = channel.getDataSet(Member.class);
    workerSet = select(workerSet, new Member.WorkerSelector(), null);
    if (Util.notEmpty(workerSet)) {
      return workerSet.size();
    }
    return 0;
  }

  /**
   * @return the regexp that contains all the path to ignore when copying a webapp.
   * @since jcms-6.0.1
   */
  public static String getIgnoreFileRegexp() {
    if (ignoreFileRegexp != null) {
      return ignoreFileRegexp;
    }

    ignoreFileRegexp = channel.getProperty("deploy-mgr.ignore-file");
    String copyIgnore = channel.getProperty("deploy-mgr.copy-ignore", null);
    if (Util.notEmpty(copyIgnore)) {
      ignoreFileRegexp += "|" + copyIgnore;
    }

    return ignoreFileRegexp;
  }

  /**
   * @return the regexp that contains all the path to ignore when signing a webapp.
   * @since jcms-6.0.1
   */
  public static String getIgnoreSignRegexp() {
    if (ignoreSignRegexp != null) {
      return ignoreSignRegexp;
    }

    StringBuffer sb = new StringBuffer(1024);
    sb.append(getIgnoreFileRegexp()).append("|").append(channel.getProperty("deploy-mgr.ignore-sign"));

    ignoreSignRegexp = sb.toString();
    return ignoreSignRegexp;
  }

  /**
   * Returns a friendly textual representation for the given date (e.g. "15 min ago", "today", "yesterday at 8:12 PM", "tomorrow at 8:12 PM", ...)
   * 
   * @param date the date
   * @param style the date and time format style (e.g. DateFormat.SHORT)
   * @param showTime if true show the time
   * @param locale the locale
   * @return a friendly textual representation for the given date.
   * @since jcms-6.0.1
   */
  public static String getFriendlyDate(Date date, int style, boolean showTime, Locale locale) {
    return getFriendlyDate(date, style, showTime, locale, false);
  }

  /**
   * Returns a friendly textual representation for the given date (e.g. "15 min ago", "today", "yesterday at 8:12 PM", "tomorrow at 8:12 PM", ...)
   * 
   * @param date the date
   * @param style the date and time format style (e.g. DateFormat.SHORT)
   * @param showTime if true show the time
   * @param locale the locale
   * @param beginSentence if true returns a string suitable for a sentence beginning.
   * @return a friendly textual representation for the given date.
   * @since jcms-6.0.1
   */
  public static String getFriendlyDate(Date date, int style, boolean showTime, Locale locale, boolean beginSentence) {

    if (date == null) {
      return "";
    }

    String lang = I18nUtil.getLanguageKey(locale);

    String timeStr = showTime ? DateFormat.getTimeInstance(style, locale).format(date) : "";

    long duration = date.getTime() - System.currentTimeMillis();

    String suffix = beginSentence ? ".b" : ".e";

    if (showTime) {
      // Near Futur
      if (duration > 0 && duration < MILLIS_IN_ONE_HOUR) {
        if (duration < MILLIS_IN_ONE_MINUTE) {
          return glp(lang, "ui.f-date.next-min" + suffix);
        }
        // Remove seconds
        duration = (duration / MILLIS_IN_ONE_MINUTE) * MILLIS_IN_ONE_MINUTE;
        return glp(lang, "ui.f-date.next-hour" + suffix, DateUtil.formatDuration(duration, locale));
      }

      // Near Past
      if (duration <= 0 && -duration < MILLIS_IN_ONE_HOUR) {
        if (-duration < MILLIS_IN_ONE_MINUTE) {
          return glp(lang, "ui.f-date.last-min" + suffix);
        }
        // Remove seconds
        duration = (-duration / MILLIS_IN_ONE_MINUTE) * MILLIS_IN_ONE_MINUTE;
        return glp(lang, "ui.f-date.last-hour" + suffix, DateUtil.formatDuration(duration, locale));
      }
    }

    // Today
    if (DateUtil.isToday(date)) {
      if (showTime) {
        return glp(lang, "ui.f-date.today.time" + suffix, timeStr);
      }
      return glp(lang, "ui.f-date.today" + suffix);
    }

    // Yesterday
    if (DateUtil.isYesterday(date)) {
      if (showTime) {
        return glp(lang, "ui.f-date.yesterday.time" + suffix, timeStr);
      }
      return glp(lang, "ui.f-date.yesterday" + suffix);
    }

    // Tomorrow
    if (DateUtil.isTomorrow(date)) {
      if (showTime) {
        return glp(lang, "ui.f-date.tomorrow.time" + suffix, timeStr);
      }
      return glp(lang, "ui.f-date.tomorrow" + suffix);
    }

    String dateStr = null;

    // This week
    if (duration > 0 && duration < 7 * MILLIS_IN_ONE_DAY) {
      dateStr = glp("ui.f-date.thisweek" + suffix, new SimpleDateFormat(glp(lang, "ui.f-date.thisweek.pattern"), locale).format(date), timeStr);
      if (beginSentence) {
        dateStr = Util.recapitalize(dateStr);
      }

    }

    // A long time
    else {
      dateStr = DateFormat.getDateInstance(style, locale).format(date);
      if (beginSentence) {
        dateStr = Util.recapitalize(dateStr);
      }
    }
    if (showTime) {
      return glp(lang, "ui.f-date.long-ago.time" + suffix, dateStr, timeStr);
    }

    return glp(lang, "ui.f-date.long-ago" + suffix, dateStr);

  }

  /**
   * Returns a friendly textual representation of a time period defined by two dates.
   * 
   * @param startDate the start date (must not be null)
   * @param endDate the end date (may be null)
   * @param locale the locale
   * @return a friendly textual representation of a time period defined by two dates.
   * @since jcms-6.1.0
   */
  public static String getFriendlyPeriod(Date startDate, Date endDate, Locale locale) {

    if (startDate == null) {
      return "";
    }

    if (locale == null) {
      locale = channel.getLocale();
    }

    String lang = I18nUtil.getLanguageKey(locale);

    SimpleDateFormat monthYearFormat = new SimpleDateFormat("MMM yyyy", locale);

    if (endDate == null) {
      return glp(lang, "ui.f-period.since", monthYearFormat.format(startDate));
    }

    if (startDate.after(endDate)) {
      return "";
    }

    SimpleDateFormat format = new SimpleDateFormat("yyyy", locale);
    long delta = endDate.getTime() - startDate.getTime();
    if (delta < 9 * MILLIS_IN_ONE_MONTH) {
      format = monthYearFormat;
    }

    String str = format.format(startDate);
    if (delta > MILLIS_IN_ONE_MONTH) {
      str = str + " - " + format.format(endDate);
    }
    return str;

  }

  // pattern used to find all relative url (href or src) and replace them with the base url provided if any
  static Pattern relativeUrlPattern;
  static Pattern absoluteUrlAndMailPattern; // used to check if the url is eitheir an an email link or an absolute url

  static Perl5Compiler regexpCompiler = new Perl5Compiler();

  static {
    try {
      // pattern used to find all relative url (href or src) and replace them with the base url provided if any
      relativeUrlPattern = regexpCompiler.compile("((?:href|src)=[\"'])([^\"']*)([\"'])");
      // used to check if the url is eitheir an an email link or an absolute url
      absoluteUrlAndMailPattern = regexpCompiler.compile("(?:(?:\\w{2,5}://)|(?:mailto:)).*");
    } catch (org.apache.oro.text.regex.MalformedPatternException ex) {
      logger.warn("An exception occured while compiling wiki perl5 pattern", ex);
    }
  }

  /**
   * Convert all relative URI in the given text to complete URL with the given base URL.
   * 
   * @param text the text to process
   * @param baseURL the base URL
   * @return the text with only complete URL
   */
  public static String convertUri2Url(String text, String baseURL) {
    StringBuffer sb = new StringBuffer();
    int offset = 0;
    Perl5Matcher absoluteUrlMatcher = new Perl5Matcher();
    PatternMatcherInput input = new PatternMatcherInput(text);

    Perl5Matcher wikiMatcher = new Perl5Matcher();
    wikiMatcher.setMultiline(true);
    while (wikiMatcher.contains(input, relativeUrlPattern)) {

      // Copy string before matching
      sb.append(input.preMatch());
      MatchResult results = wikiMatcher.getMatch();
      String grp1 = results.group(1);
      String url = results.group(2);
      String grp2 = results.group(3);

      // It is already an absolute URL, leave as is
      if (absoluteUrlMatcher.matches(url, absoluteUrlAndMailPattern)) {
        sb.append(input.match());
      }
      // It is a relative URL add the base URL.
      else {
        sb.append(grp1).append(baseURL).append(url).append(grp2);
      }

      offset = input.getMatchEndOffset();
      input.setBeginOffset(input.getMatchEndOffset());
    }
    // Copy string after last matching
    sb.append(text.substring(offset));

    return sb.toString();
  }

  /**
   * Returns the map file size for each content-type.
   *
   * @return the map file size for each content-type.
   * @since jcms-7.0.0
   * @deprecated since jcms-7.1.0, use Channel.getGlobalFileAnalyzer method to have more information about satellites files
   */
  public static ObjectLongTreeMap getFileSizeMap() {
    return QuotaManager.getInstance().getGlobalFileAnalyzer(false).getFileSizeMap();
  }

  /**
   * Returns the map file size for each content-type for the given workspace.
   *
   * @param ws the workspace
   * @return the map file size for each content-type.
   * @since jcms-7.0.0
   * @deprecated since jcms-7.1.0, use Workspace.getFileAnalyzer method to have more information about satellites files
   */
  public static ObjectLongTreeMap getFileSizeMap(Workspace ws) {
    return ws.getFileAnalyzer(false).getFileSizeMap();
  }

  /**
   * Remove from a value characters which should not be in email headers (like : or ,)
   * 
   * @param value the value to check
   * @return the value without problematic characters
   */
  public static String getEscapedEmailName(String value) {
    return value == null ? null : Util.replace(value, "s/[:,]//g");
  }

  // ---------------------------------------------------------------------------------------------------------------
  // HTML Helper
  // ---------------------------------------------------------------------------------------------------------------

  public static String computeName(String src) {
    String name = src;
    int p1 = src.lastIndexOf('/');
    int p2 = src.lastIndexOf('.');
    if (p1 >= 0 && p2 >= 0) {
      name = src.substring(p1 + 1, p2);
    }
    return name;
  }

  /**
   * Build an icon
   * 
   * @param src the image source
   * @return the image tag representation
   * @since jcms-9.0.0
   */
  public static String getHtmlIcon(String src) {
    return getHtmlIcon(src, null, null, null);
  }

  /**
   * Build an icon
   * 
   * @param src the image source
   * @param alt the image alt
   * @return the image tag representation
   * @since jcms-9.0.0
   */
  public static String getHtmlIcon(String src, String alt) {
    return getHtmlIcon(src, alt, null, null);
  }

  /**
   * Build an icon
   * 
   * @param src the image source
   * @param alt the image alt
   * @param title the image title
   * @param css the css for <code>class=""</code> attribute.
   * @return the image tag representation
   * @since jcms-9.0.0
   */
  public static String getHtmlIcon(String src, String alt, String title, String css) {
    return getHtmlIcon(src, alt, title, css, null);
  }

  /**
   * Build an icon
   * 
   * @param src the image source
   * @param alt the image alt
   * @param title the image title
   * @param css the css for <code>class=""</code> attribute.
   * @param htmlAttributes the html attributes
   * @return the image tag representation
   * @since jcms-9.0.0
   */
  public static String getHtmlIcon(String src, String alt, String title, String css, String htmlAttributes) {
    return getHtmlIcon(src, alt, title, css, htmlAttributes, null);
  }

  /**
   * Build an icon
   * 
   * @param src the image source
   * @param alt the image alt
   * @param title the image title
   * @param customCSS the css for <code>class=""</code> attribute.
   * @param htmlAttributes the html attributes
   * @param lang the user language
   * @return the image tag representation
   * @since jcms-9.0.0
   */
  public static String getHtmlIcon(String src, String alt, String title, String customCSS, String htmlAttributes, String lang) {
    return getHtmlIcon(src, alt, title, customCSS, htmlAttributes, lang, 0);
  }

  private static String getHtmlIcon(String src, String alt, String title, String customCSS, String htmlAttributes, String lang, int recursive) {

    if (recursive > 5) {
      throw new IllegalArgumentException("Too many recursive calls.");
    }

    // Remove the icon prefix if found
    if (src != null && src.startsWith(JcmsUtil.ICON_PREFIX_PROP)) {
      src = src.substring(ICON_PREFIX_PROP.length());
    }

    if (Util.isEmpty(src)) {
      return "";
    }

    String css = "";
    if (Util.notEmpty(customCSS)) {
      css = css + " " + customCSS;
    }

    if (Util.notEmpty(lang)) {
      // Try to resolve alt as a lang property
      if (alt == null) {
        alt = channel.getProperty(lang + "." + JcmsUtil.ICON_PREFIX_PROP + src, null);
      } else if (Util.notEmpty(alt)) {
        alt = glp(lang, alt);
      }

      // Try to resolve title as lang prop
      if (Util.notEmpty(title)) {
        title = glp(lang, title);
      }
    }

    // Try to resolve src as a property
    String realSrc = channel.getProperty(src, null);
    if (realSrc == null) {
      realSrc = channel.getProperty(ICON_PREFIX_PROP + src, src);
    }
    if (!realSrc.equals(src)) {
      css = css + " " + src;
    }

    // Glyph
    if (realSrc.startsWith(GLYPH_ICON_PREFIX)) {
      String glyph = realSrc.substring(GLYPH_ICON_PREFIX.length());
      css += " " + glyph.trim();
      return HtmlUtil.getSpanIcon(css.trim(), alt, title, htmlAttributes);
    }

    // Sprite
    if (realSrc.startsWith(SPRITE_ICON_PREFIX)) {
      String sprite = realSrc.substring(SPRITE_ICON_PREFIX.length());
      css += " " + sprite.trim();
      return HtmlUtil.getHtmlIcon("s.gif", alt, title, css.trim(), htmlAttributes);
    }

    // Icon Property
    if (realSrc.startsWith(ICON_ICON_PREFIX)) {
      String prop = realSrc.substring(ICON_ICON_PREFIX.length()).trim();
      return getHtmlIcon(prop, alt, title, customCSS, htmlAttributes, lang, recursive + 1);
    }

    // Path
    return HtmlUtil.getHtmlIcon(realSrc, alt, title, css.trim(), htmlAttributes);
  }

  /**
   * Instantiate new objects according to a property key and the given {@link Class} representing the type to return.<br>
   * It is possible to define several classes to be instantiated by separating them by a whitespace, a comma, or a line break.<br>
   * Examples:<br>
   * <code>com.my.key: com.jalios.xxx.xxx.MyServiceClass</code><br>
   * <code>com.my.key: com.jalios.xxx.xxx.MyServiceClass,com.jalios.xxx.xxx.MyOtherServiceClass</code>
   *
   * @param clazz {@link Class} representing the {@link Class} to return
   * @param propKey property key where find the {@link Class} name(s) (eg. com.my.key: com.jalios.xxx.xxx.MyServiceClass with 'com.my.key' is
   *          <b>propKey</b>)
   * @return Objects list of type <b>clazz</b>
   */
  public static <T> List<T> createInstances(Class<T> clazz, String propKey) {
    String classNames = channel.getProperty(propKey, null);

    if (Util.isEmpty(classNames)) {
      logger.warn("No property '" + propKey + "' was found.");
      return null;
    }

    List<T> list = new ArrayList<T>();

    StringTokenizer st = new StringTokenizer(classNames, DELIMITERS);
    while (st.hasMoreTokens()) {
      String className = st.nextToken();
      className = className.trim();
      if (className.length() > 0) {
        try {
          Class<T> c = (Class<T>) channel.getClass(className);
          T obj = c.newInstance();
          logger.debug("Object initialized: " + c.getCanonicalName());
          list.add(obj);
        } catch (ClassNotFoundException e) {
          logger.warn("Invalid classname '" + classNames + "' for the property '" + propKey + "': " + e.getMessage(), e);
        } catch (InstantiationException e) {
          logger.warn("Invalid classname '" + classNames + "' for the property '" + propKey + "': " + e.getMessage(), e);
        } catch (IllegalAccessException e) {
          logger.warn("Invalid classname '" + classNames + "' for the property '" + propKey + "': " + e.getMessage(), e);
        }
      }
    }

    return list;
  }

  /**
   * Instantiate new objects according to property keys starting with the given prefix and the given {@link Class} representing the type to
   * return.<br>
   * It is possible to define several classes to be instantiated by separating them by a whitespace, a comma, or a line break.<br>
   * Examples:<br>
   * <code>com.my.key.1: com.jalios.xxx.xxx.MyServiceClass</code><br>
   * <code>com.my.key.2: com.jalios.xxx.xxx.MyOtherServiceClass</code><br>
   * <br>
   * <code>com.my.key.1: com.jalios.xxx.xxx.MyServiceClass</code><br>
   * <code>com.my.key.2: com.jalios.xxx.xxx.MySecondServiceClass,com.jalios.xxx.xxx.MyThirdServiceClass</code>
   *
   * @param clazz {@link Class} representing the {@link Class} to return
   * @param propertyPrefix property prefix used to find the property keys related to the {@link Class} name(s)
   * @return Objects list of type <b>clazz</b>
   */
  public static <T> List<T> createInstancesFromPropertyPrefix(Class<T> clazz, String propertyPrefix) {
    Map<String, String> map = channel.getProperties(propertyPrefix);
    if (map == null) {
      logger.warn("No properties with prefix '" + propertyPrefix + "' was found.");
      return null;
    }

    List<T> list = new ArrayList<T>();

    for (Map.Entry<String, String> entry : map.entrySet()) {
      String key = entry.getKey();
      String value = entry.getValue();
      List<T> l = createInstances(clazz, key);
      if (l != null) {
        list.addAll(l);
      }
    }

    return list;
  }

}
