/*
 * ------------------------------------------------------------------------
 * Copyright (c) Jalios 2002.
 * DATE AUTHOR CHANGES
 * 2003/03/26 O. Dedieu Created
 * ------------------------------------------------------------------------
 */

package com.jalios.jcms.taglib;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.jsp.JspException;
import javax.servlet.jsp.JspWriter;
import javax.servlet.jsp.PageContext;
import javax.servlet.jsp.tagext.BodyContent;
import javax.servlet.jsp.tagext.BodyTagSupport;

import org.apache.log4j.Logger;

import com.jalios.jcms.Channel;
import com.jalios.jcms.Form;
import com.jalios.jcms.JcmsUtil;
import com.jalios.jcms.Publication;
import com.jalios.jcms.ajax.AjaxRequestWrapper;
import com.jalios.jcms.context.JcmsJspContext;
import com.jalios.jcms.plugin.PluginManager;
import com.jalios.jcms.portlet.PortalElement;
import com.jalios.jcms.portlet.PortalManager;
import com.jalios.jtaglib.BufferTag;
import com.jalios.util.Util;

public class IncludeTag extends BodyTagSupport {

  public static final String REVISION = "$Revision: 97819 $";

  private static final long serialVersionUID = 1L;

  private static final Logger LOG = Logger.getLogger(IncludeTag.class);

  public static final String INCLUDE_BODY = "INCLUDE_BODY";
  public static final String INCLUDE_CONTEXT = "INCLUDE_CONTEXT";
  public static final String JCMS_RESOURCE_INCLUDE_PREFIX_PROP = "jcms.resource.include.";

  private boolean injectContent = true;

  private String file;
  private String jsp;
  private String clazz;
  private String id;
  private Publication pub;
  private String target;
  private String targetContext;
  private String usage;
  private String usageBackup;
  private String resource;

  private boolean included;

  // ----------------------------------------------------------------------
  // Static Helper
  // ----------------------------------------------------------------------

  /**
   * Helper to store a value like a &lt;jalios:buffer&gt; in an include context to override some values
   * 
   * @param request the current Request
   * @param name the name to store
   * @param value the value associated
   */
  public static void setIncludeString(ServletRequest request, String name, String value) {
    Map<String, Object> bufferMap = (Map<String, Object>) request.getAttribute(BufferTag.BUFFER_CONTEXT);
    if (bufferMap == null) {
      bufferMap = new HashMap<String, Object>();
      request.setAttribute(BufferTag.BUFFER_CONTEXT, bufferMap);
    }
    bufferMap.put(name, value);
  }

  /**
   * Retrieve String from contextual include
   * 
   * @param request the current Request
   * @param name the name to store
   * @param def the default value
   * @return String the associated String
   */
  public static String getIncludeString(ServletRequest request, String name, String def) {
    Object obj = getIncludeObject(request, name, def);
    if (obj == null) {
      return def;
    }

    return Util.getString(obj, def);
  }

  /**
   * Retrieve Object from contextual include
   * 
   * @param request the current Request
   * @param name the name to store
   * @param def the default value
   * @return Object the associated Object
   */
  public static Object getIncludeObject(ServletRequest request, String name, Object def) {
    if (name == null || request == null) {
      return def;
    }
    Map<String, Object> bufferMap = (Map<String, Object>) request.getAttribute(IncludeTag.INCLUDE_CONTEXT);
    if (bufferMap == null) {
      return def;
    }
    Object tmp = bufferMap.get(name);
    return tmp != null ? tmp : def;
  }

  // ----------------------------------------------------------------------
  // Attribute setters
  // ----------------------------------------------------------------------

  public String getFile() {
    return file;
  }

  public void setFile(String file) {
    this.file = file;
  }

  public String getJsp() {
    return jsp;
  }

  public void setJsp(String jsp) {
    this.jsp = jsp;
  }

  public String getResource() {
    return resource;
  }

  public void setResource(String resource) {
    this.resource = resource;
  }

  public String getClazz() {
    return clazz;
  }

  public void setClazz(String clazz) {
    this.clazz = clazz;
  }

  @Override
  public String getId() {
    return id;
  }

  @Override
  public void setId(String id) {
    this.id = id;
  }

  public Publication getPub() {
    return pub;
  }

  public void setPub(Publication pub) {
    this.pub = pub;
  }

  public String getTarget() {
    return target;
  }

  public void setTarget(String target) {
    this.target = target;
  }

  public String getTargetContext() {
    return targetContext;
  }

  public void setTargetContext(String targetContext) {
    this.targetContext = targetContext;
  }

  public String getUsage() {
    return usage;
  }

  public void setUsage(String usage) {
    this.usage = usage;
  }

  // ----------------------------------------------------------------------
  // Common Methods
  // ----------------------------------------------------------------------

  public void clean() {
    file = null;
    jsp = null;
    clazz = null;
    id = null;
    pub = null;
    target = null;
    targetContext = null;
    usage = null;
    usageBackup = null;
    resource = null;
    backupBufferMap = null;
  }

  protected void setContext(Map<String, Object> map, String body) {
    if (body != null) {
      map.put(INCLUDE_BODY, body);
    }
  }

  // ----------------------------------------------------------------------
  // Tag handlers
  // ----------------------------------------------------------------------

  Map<String, Object> backupBufferMap = null;

  @Override
  public int doStartTag() throws JspException {
    Map<String, Object> bckpBufferMap = (Map<String, Object>) pageContext.getRequest().getAttribute(BufferTag.BUFFER_CONTEXT);
    if (bckpBufferMap != null) {
      backupBufferMap = new HashMap<String, Object>(bckpBufferMap);
    }
    return EVAL_BODY_BUFFERED;
  }

  @Override
  public void doInitBody() throws JspException {
    doBeforeBody();
    super.doInitBody();
  }

  @Override
  public int doAfterBody() throws JspException {

    BodyContent body = getBodyContent();
    String content = body.getString();

    // Clear body
    // @formatter:off
    try { body.clear(); } catch (IOException ex) { /* empty */ }
    // @formatter:on

    // Perform includes
    include(content);

    // Inject content
    if (injectContent) {
      try {
        body.writeOut(body.getEnclosingWriter());
      } catch (IOException ex) {
        printErrorMessage(jsp, clazz, ex, pageContext);
      }
    }

    // Clear body
    // @formatter:off
    try { body.clear(); } catch (IOException ex) { /* empty */ }
    // @formatter:on

    return SKIP_BODY;
  }

  @Override
  public int doEndTag() {
    include();
    included = false;
    return EVAL_PAGE;
  }

  /**
   * Called on Init Body and for each body loops
   */
  protected void doBeforeBody() {
  }

  // ----------------------------------------------------------------------
  // INCLUDES
  // ----------------------------------------------------------------------

  /**
   * Check if the current publication or the current portal element
   * is a children of the given publication to include (Using TreeNode Hierarchy)
   * 
   * @param pub the publication to include
   * @return boolean false if it is a recursive include.
   */
  protected boolean notRecusiveInclude(Publication pub) {
    PortalElement currentPE = (PortalElement) pageContext.getRequest().getAttribute("PortalElement");
    if (currentPE == null) {
      return true;
    }

    if (Util.getTreeDescendantSet(pub).contains(currentPE)) {
      LOG.warn("Recursive Include of PortalElement: " + currentPE);
      return false;
    }
    return true;
  }

  protected void include() {
    include(null);
  }

  protected void include(String body) {
    if (included) {
      return;
    }
    included = true;

    // [Context] Build include context
    ServletRequest req = pageContext.getRequest();
    Object backupInclude = req.getAttribute(IncludeTag.INCLUDE_CONTEXT);

    Map<String, Object> includeMap = new HashMap<String, Object>();
    req.setAttribute(IncludeTag.INCLUDE_CONTEXT, includeMap);
    if (Util.notEmpty(resource)) {
      includeMap.put("resource", resource);
    }

    Map<String, Object> bufferMap = (Map<String, Object>) req.getAttribute(BufferTag.BUFFER_CONTEXT);

    // [Context] Store tag's body
    setContext(includeMap, body);

    // [Buffer] Store buffers

    if (bufferMap != null) {
      includeMap.putAll(bufferMap);
    }
    req.removeAttribute(BufferTag.BUFFER_CONTEXT);

    // [Include] Retrieve JcmsContext
    JcmsJspContext ctxt = (JcmsJspContext) req.getAttribute("jcmsContext");
    if (ctxt.getPageContext() != pageContext) {
      ctxt = new JcmsJspContext(pageContext);
    }

    // [Include] Set the usage for this include
    if (usage != null) {
      usageBackup = ctxt.getTemplateUsage();
      ctxt.setTemplateUsage(usage);
    }

    try {
      // [Include] Include a File
      if (file != null) {
        includeFile(file);
        return;
      }

      // [Include] Include a Target
      if (target != null) {
        includeTarget(req, target);
        return;
      }

      // [Include] Check Parameter Id or pub
      if (id != null) {
        pub = Channel.getChannel().getPublication(id);
      }
      jsp = Util.getString(includePub(ctxt, pub), jsp);

      // [Include] Check parameter clazz
      jsp = Util.getString(includeClass(clazz), jsp);

      if (Util.notEmpty(resource) && Util.isEmpty(jsp)) {
        String resourceKey = JCMS_RESOURCE_INCLUDE_PREFIX_PROP + resource;
        jsp = Channel.getChannel().getProperty(resourceKey, null);
        if (Util.isEmpty(jsp)) {
          LOG.warn("The resource '" + resourceKey + "' could not be found.");
        }
      }

      if (Util.notEmpty(jsp)) {
        JcmsJspContext jcmsContextBackup = null;
        boolean hasParam = jsp.contains("?"); 
        if (hasParam) {
          // Force doInitPage.jspf to recompute a new JcmsJspContext based on the new ServletRequest
          // that will be built by app server for the include() operation.
          //
          // This approach is needed to ensure that any methods relying on the channel.getCurrentServletRequest
          // jsp correctly reads any parameter that might be specified through this tag <jalios:include jsp="path/to.jsp?foo=true"/>.
          //
          // Without this "hack", JSP being invoked might get different parameter value depending which 
          // method is invoked (ie request object being used : "original" request vs "included" request)
          //   for example : request.getParameter("foo") != getBooleanParameter("foo", false)
          // 
          // TODO Improvement : the cleanest way to perform this "request update" would be to  
          // systematically use a HttpServletRequestWrapper, and implement the setRequest() method
          // to be informed of request change during include/forward
          // more info https://wiki.magnolia-cms.com/display/DEV/How+servlet+request+wrapping+works
          //
          req.setAttribute("forceJcmsContextUpdate", Boolean.TRUE);
          jcmsContextBackup = (JcmsJspContext) req.getAttribute("jcmsContext");
          
          // Disable the AjaxRequestWrapper if any parameter is specified in the JSP path
          // to ensure that any javabean handler are correctly invoked with those parameters,
          // and not only with the initial parameters provided by the AjaxRequestWrapper
          AjaxRequestWrapper.disable(req);
        }
        
        pageContext.include("/" + jsp.trim());
        
        if (hasParam) {
          // Restore default behavior
          AjaxRequestWrapper.enable(req);
          
          // restore previous context and current servlet request
          req.removeAttribute("forceJcmsContextUpdate");
          req.setAttribute("jcmsContext",jcmsContextBackup);
          Channel.getChannel().setCurrentServletRequest((HttpServletRequest) req);
        }
        
      }
      
    } catch (ClassNotFoundException ex) {
      printErrorMessage(jsp, clazz, ex, pageContext);
    } catch (ServletException ex) {
      printErrorMessage(jsp, clazz, ex, pageContext);
    } catch (IOException ex) {
      printErrorMessage(jsp, clazz, ex, pageContext);
    } finally {

      // [Context] clean stuff
      if (backupInclude != null) {
        req.setAttribute(IncludeTag.INCLUDE_CONTEXT, backupInclude);
      } else {
        req.removeAttribute(IncludeTag.INCLUDE_CONTEXT);
      }

      // [Buffer] clean stuff
      if (backupBufferMap != null) {
        Map<String, Object> localBufferMap = (Map<String, Object>) req.getAttribute(BufferTag.BUFFER_CONTEXT);
        if (localBufferMap != null) {
          backupBufferMap.putAll(localBufferMap);
        }
        req.setAttribute(BufferTag.BUFFER_CONTEXT, backupBufferMap);
      }

      // [Include] Restore template usage
      if (usage != null) {
        ctxt.setTemplateUsage(usageBackup);
      }
      clean();
    }
    return;
  }

  private void includeFile(String file) throws IOException {
    InputStreamReader reader = null;
    try {
      File theFile = new File(Channel.getChannel().getRealPath(file));
      reader = new InputStreamReader(new FileInputStream(theFile), pageContext.getResponse().getCharacterEncoding());
      JspWriter writer = pageContext.getOut();
      char[] buf = new char[1024];
      for (int c; (c = reader.read(buf)) != -1;) {
        writer.write(buf, 0, c);
      }
    } finally {
      if (reader != null) {
        reader.close();
      }
    }
  }

  private void includeTarget(ServletRequest req, String target) {
    req.setAttribute("IncludeTag", target);
    displayTarget(target, targetContext, pageContext);
    if (Util.notEmpty(targetContext)) {
      req.setAttribute("IncludeTagTargetContext", targetContext);
    }
    PluginManager.getPluginManager().performInclude(target, pageContext);
    if (Util.notEmpty(targetContext)) {
      req.removeAttribute("IncludeTagTargetContext");
    }
    req.removeAttribute("IncludeTag");
  }

  // FIXME: [Template] Should use template declaration
  private String includeClass(String clazz) throws ClassNotFoundException {
    if (clazz == null) {
      return null;
    }

    Class<?> c = Channel.getChannel().getClass(clazz);
    if (!Form.class.isAssignableFrom(c)) {
      return null;
    }
    String form = Util.getClassShortName(clazz);
    return "types/" + form + "/editForm" + form + ".jsp";
  }

  private String includePub(JcmsJspContext ctxt, Publication pub) {
    if (pub == null) {
      return null;
    }
    if (!notRecusiveInclude(pub)) {
      return null;
    }

    ServletRequest req = ctxt.getRequest();

    if (pub instanceof PortalElement) {
      req.setAttribute("displayPortlet", pub);
      req.setAttribute(PortalManager.PORTAL_TEMPLATEUSAGE, usage);
      return "jcore/portal/doDynIncludePortlet.jsp";
    }

    req.setAttribute(PortalManager.PORTAL_PUBLICATION, pub);
    return pub.getTemplatePath(ctxt);
  }

  protected void skipInclude() {
    included = true;
  }

  // ----------------------------------------------------------------------
  // UTILITY
  // ----------------------------------------------------------------------

  /**
   * Print an Exception message in log and in JSP (if developement mode)
   * 
   * @param jsp the JSP used
   * @param clazz the class used
   * @param ex the exception
   * @param pageContext the given pageContext
   */
  public static void printErrorMessage(String jsp, String clazz, Exception ex, PageContext pageContext) {
    try {
      PortalManager.printPortletException(pageContext.getOut(), ex, jsp, LOG);
    } catch (IOException ex2) {
      LOG.warn("printErrorMessage: an exception occured: ", ex2);
    }
  }

  private static void displayTarget(String target, String targetContext, PageContext pageContext) {
    if (!Util.toBoolean(pageContext.getSession().getAttribute("channel.dev-tools.targets"), false)) {
      return;
    }

    JspWriter out = pageContext.getOut();

    boolean inTable = false;
    boolean inHeader = false;
    boolean inList = false;
    boolean inSkin = false;
    boolean inTRow = false;
    boolean inDiv = false;
    boolean inSpan = false;
    boolean inTabPage = false;
    boolean isEmpty = false;
    boolean inHtmlComments = false;

    if (Util.notEmpty(targetContext)) {
      String[] targetContexts = targetContext.split("\\s");
      List<String> targetContextsSet = Arrays.asList(targetContexts);

      inTable = targetContextsSet.contains("tr");
      inHeader = targetContextsSet.contains("header");
      inList = targetContextsSet.contains("li");
      inSkin = targetContextsSet.contains("skin");
      inTRow = targetContextsSet.contains("td");
      inDiv = targetContextsSet.contains("div") || targetContextsSet.contains("bloc");
      inSpan = targetContextsSet.contains("span") || targetContextsSet.contains("inline");
      inTabPage = targetContextsSet.contains("tab-page");
      isEmpty = targetContextsSet.contains("empty");
      inHtmlComments = targetContextsSet.contains("htmlcomments");
    }
    if (isEmpty) {
      return;
    }

    try {
      if (inSkin) {
        if (Channel.getChannel().getBooleanProperty("channel.dev-tools.targets.fo", false)) {
          // @formatter:off
          if (inHeader || inHtmlComments){ out.println("<!--");       }
          if (inTable) { out.println("<tr><td>");   }
          if (inTRow)  { out.println("<td nowrap='nowrap'>");       } 
          if (inList)  { out.println("<li>");       }
          if (inDiv)   { out.println("<div>");       }
          if (inSpan)  { out.println("<span>");      }
          String title = target + ( Util.notEmpty(targetContext) ? "(" + targetContext + ')' : "" );
          out.println(JcmsUtil.getHtmlIcon("target", target, title, null));
          if (inSpan)  { out.println("</span>");      }
          if (inDiv)   { out.println("</div>");      }
          if (inList)  { out.println("</li>");      } 
          if (inTRow)  { out.println("</td>");      } 
          if (inTable) { out.println("</td></tr>"); } 
          if (inHeader || inHtmlComments){ out.println("-->");        }
          // @formatter:on
        }
        return;
      }

      // @formatter:off
      if (inHeader || inHtmlComments){ out.println("<!--");       }
      if (inTable) { out.println("<tr><td>");   }
      if (inTRow)  { out.println("<td nowrap='nowrap'>");       } 
      if (inList)  { out.println("<li>");       }
      if (inDiv)   { out.println("<div>");       }
      if (inSpan)  { out.println("<span>");      }
      if (inTabPage) { out.println("<div class='tab-page'><h2 class='tab'>"); }                                   
      
      out.print(JcmsUtil.getHtmlIcon("target", null));
      out.print("<span style='font: bold 7pt Verdana,sans-serif; color: #CC0000;' " + ( Util.notEmpty(targetContext) ? "title='targetContext=(" + targetContext + ")'" : "" ) + ">" + target + "</span>");
      out.print(JcmsUtil.getHtmlIcon("target", null));
  
      if (inTabPage) { 
        out.print("</h2><div class='tabinfo'>");
        out.print(JcmsUtil.getHtmlIcon("target", null));
        out.print("<span style='font: bold 7pt Verdana,sans-serif; color: #CC0000;' >" + target + "</span>");
        out.print(JcmsUtil.getHtmlIcon("target", null));
        out.print("</div></div>"); 
      }
      if (inSpan)  { out.println("</span>");      }
      if (inDiv)   { out.println("</div>");      }
      if (inList)  { out.println("</li>");      } 
      if (inTRow)  { out.println("</td>");      } 
      if (inTable) { out.println("</td></tr>"); } 
      if (inHeader || inHtmlComments) { out.println("-->");        }
      // @formatter:on
    } catch (IOException ioe) {
      LOG.debug("Exception while displaying target '" + target + "'", ioe);
    }
  }
}
