package com.jalios.jcms.taglib.settings.helper;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import javax.servlet.http.HttpServletRequest;

import org.apache.log4j.Logger;

import com.jalios.jcms.Channel;
import com.jalios.jcms.DBFileDocument;
import com.jalios.jcms.Data;
import com.jalios.jcms.FileDocument;
import com.jalios.jcms.JcmsConstants;
import com.jalios.jcms.JcmsUtil;
import com.jalios.jcms.Member;
import com.jalios.jcms.Publication;
import com.jalios.jcms.TypeEntry;
import com.jalios.jcms.TypeTemplateEntry;
import com.jalios.jcms.taglib.settings.BasicSettings;
import com.jalios.jcms.taglib.settings.ChooserConstants;
import com.jalios.jcms.taglib.settings.impl.FileSettings;
import com.jalios.jcms.taglib.settings.impl.PublicationSettings;
import com.jalios.jcms.workspace.Workspace;
import com.jalios.util.URLUtils;
import com.jalios.util.Util;

public class PublicationControlHelper extends AbstractControlHelper {

  private static final Logger LOG = Logger.getLogger(PublicationControlHelper.class);

  private boolean fileInputAdded;
  private boolean fileInputHidden;

  private boolean dropdownAdded;

  private String customCss;

  private Set<?> subTypeSet;

  public PublicationControlHelper(Member loggedMember, String userLang, Workspace workspace, HttpServletRequest request) {
    super(loggedMember, userLang, workspace, request);
    this.customCss = getOption(BasicSettings.CSS, "");
    initFileInput();
    initDropdown();
  }

  private void initFileInput() {
    String superType = getOption(PublicationSettings.SUPER_TYPE, null);
    if (superType != null) {
      Class<? extends Data> superTypeClass = null;
      try {
        superTypeClass = getOption(PublicationSettings.SUPER_TYPE_CLASS, Channel.getChannel().getClass(superType));
      } catch (ClassNotFoundException e) {
        LOG.warn("Could not resolve the super type class '" + superType + "'", e);
      }
      if (superTypeClass != null) {
        this.subTypeSet = getWorkspace().getSubTypeSet(superTypeClass, getLoggedMember());

        TypeEntry typeEntry = Channel.getChannel().getTypeEntry(superTypeClass);
        if (typeEntry != null) {
          // Resolve the chooser title
          String superTypeLabel = typeEntry.getLabel(getUserLang());
          if (Util.notEmpty(superTypeLabel)) {
            String pubChooserName = getOption(ChooserConstants.CHOOSER_NAME, "publication");
            String title = JcmsUtil.glp(getUserLang(), ChooserConstants.CHOOSER_PREFIX_PROP + pubChooserName + ".supertype.title", superTypeLabel);
            addOption(ChooserConstants.CHOOSER_TITLE, title);
          }
        }

        // Add 'upload' chooser button when the publication type is assignable from FileDocument
        if (FileDocument.class.isAssignableFrom(superTypeClass)) {
          if (!getOption(PublicationSettings.FILE_INPUT_DISABLED, false) || getOption(PublicationSettings.FILE_INPUT_ENABLED, false)) {
            // TODO Condition to remove to JCMS 10 : cf. JCMS-4920
            if ((subTypeSet != null && (subTypeSet.contains(FileDocument.class) || subTypeSet.contains(DBFileDocument.class)))
                || getOption(PublicationSettings.FILE_INPUT_ENABLED, false)) {
              // The file input must be added
              this.fileInputAdded = true;
              this.fileInputHidden = Util.notEmpty(resolveValue(getFieldValue()));

              if (!fileInputHidden) { // Hide chooser input
                addOption(BasicSettings.CSS, "hide" + (Util.notEmpty(customCss) ? " " + customCss : ""));
                includeLocalObject("INPUT_DISABLED", true);
              }
            }
          } // Check isFileInputDisabled
        }
      }
    } else {
      this.subTypeSet = getWorkspace().getSubTypeSet(Publication.class, getLoggedMember());
    }
  }

  private void initDropdown() {
    // Check if the dropdown is disabled or not. If true, we don't display a dropdown menu but just a clear button
    if (!Util.toBoolean(getOption(PublicationSettings.DROPDOWN_DISABLED, null), false)) {
      // The dropdown must be added
      dropdownAdded = true;
    }
  }

  public boolean isFileInputAdded() {
    return fileInputAdded;
  }

  public boolean isFileInputHidden() {
    return fileInputHidden;
  }

  public FileSettings getFileSettings() {
    FileSettings.Mode mode = Util.toBoolean(getRequest().getAttribute(JcmsConstants.AJAX_MODE_ATTR), false) ? FileSettings.Mode.PLUPLOAD
        : FileSettings.Mode.SIMPLE_FILE;

    FileSettings settings = new FileSettings().name(getFieldName()).mode(mode).singleFile();

    if (isFieldDisabled() || isFileInputHidden()) {
      settings.disable();
    }

    if (isFieldRequired()) {
      settings.require();
    }

    String css = " jalios-" + getControlType().toString().toLowerCase() + "-filebrowser";
    css += FileSettings.Mode.PLUPLOAD.equals(mode) ? " form-control" : "";
    css += Util.notEmpty(customCss) ? " " + customCss : "";
    css += isFileInputHidden() ? " hide" : "";
    settings.css(css);

    return settings;
  }

  public boolean isDropdownAdded() {
    return dropdownAdded && Util.notEmpty(subTypeSet);
  }

  public String getListItems() {
    boolean disabled = isFieldDisabled();

    StringBuilder sb = new StringBuilder();

    Data data = Channel.getChannel().getData(resolveValue(getFieldValue()));
    // Edit item
    sb.append(generateEditItem(data, data == null, disabled));
    // Add item
    sb.append(generateAddItem(subTypeSet, data != null, disabled));
    // Insert item
    if (isFieldMultivalued()) {
      sb.append("<li class=\"insert-item\">");
      sb.append("<a href=\"#\" class=\"").append(!disabled ? " insert-field" : "").append("\" onclick=\"return false;\">");
      sb.append(JcmsUtil.getHtmlIcon("add-widget")).append(" ").append(JcmsUtil.glp(getUserLang(), "ui.com.alt.insert"));
      sb.append("</a>");
      sb.append("</li>");
    }
    // Clear item
    sb.append("<li class=\"clear-item\">");
    sb.append("<a href=\"#\"");
    sb.append(" data-jalios-action=\"widget:clear\"");
    sb.append(disabled ? " disabled" : "");
    sb.append(" onclick=\"return false;\">");
    sb.append(JcmsUtil.getHtmlIcon("clear")).append(" ").append(JcmsUtil.glp(getUserLang(), "ui.com.alt.remove"));
    sb.append("</a>");
    sb.append("</li>");

    return sb.toString();
    /*
    if (data != null) {
      return generateEditItem(data, data == null, disabled);
    } else if (Util.notEmpty(subTypeSet)) {
      return generateAddItem(subTypeSet, data != null, disabled);
    }
    return "";
    */
  }

  private String generateEditItem(Data data, boolean hidden, boolean disabled) {
    // Try to get the publication to know if it exists, and so display the dropdown menu
    String editTemplatePath = (data != null && data instanceof Publication)
        ? ((Publication) data).getTemplatePath(TypeTemplateEntry.USAGE_EDIT_POPUP, false) : "edit.jsp";

    Map<String, String[]> paramsMap = new HashMap<String, String[]>();
    paramsMap.put("id", new String[] { "{value}" });
    paramsMap.put("popupEdition", new String[] { "true" });
    paramsMap.put("redirect", new String[] { "$CHOOSER$" });
    String editUrl = URLUtils.buildUrl(editTemplatePath, paramsMap);

    StringBuilder sb = new StringBuilder();
    sb.append("<li class=\"edit-item").append(hidden ? " hide" : "").append("\">");
    sb.append("<a href=\"#\"");
    sb.append(" data-jalios-action=\"chooser:edit\" data-jalios-chooser=\"").append(editUrl).append("\"");
    sb.append(disabled ? " disabled" : "");
    sb.append(" onclick=\"return false;\">");
    sb.append(JcmsUtil.getHtmlIcon("edit")).append(" ").append(JcmsUtil.glp(getUserLang(), "ui.com.alt.edit"));
    sb.append("</a>");
    sb.append("</li>");

    if (data != null && data instanceof FileDocument) {
      sb.append("<li class=\"upload-new-version-item").append(hidden ? " hide" : "").append("\">");
      sb.append("<a href=\"#\"");
      sb.append(" data-jalios-action=\"chooser:document\"");
      String options = "id=" + data.getId() + "&type=" + data.getClass().getCanonicalName() + "&" + getOption(ChooserConstants.CHOOSER_OPTION, "");
      sb.append(" data-jalios-options=\"" + options + "\"");
      sb.append(disabled ? " disabled" : "");
      sb.append(" onclick=\"return false;\">");
      sb.append(JcmsUtil.getHtmlIcon("upload")).append(" ").append(JcmsUtil.glp(getUserLang(), "ui.filedocument.lbl.update"));
      sb.append("</a>");
      sb.append("</li>");
    }

    return sb.toString();
  }

  private String generateAddItem(Set<?> typeSet, boolean hidden, boolean disabled) {
    StringBuilder sb = new StringBuilder();
    if (typeSet.size() > 1) {
      Set<Class<?>> sortedClass = JcmsUtil.getSortedClass(typeSet, getWorkspace(), getUserLang());

      sb.append("<li class=\"dropdown-submenu add-item").append(hidden ? " hide" : "").append("\">");
      sb.append("<a href=\"#\"");
      sb.append(disabled ? " disabled" : "");
      sb.append(" onclick=\"return false;\">");
      sb.append(JcmsUtil.getHtmlIcon("add")).append(" ").append(JcmsUtil.glp(getUserLang(), "ui.com.alt.add"));
      sb.append("</a>");

      sb.append("<ul class=\"dropdown-menu\">");
      for (Object obj : sortedClass) {
        Class<? extends Data> type = (Class<? extends Data>) obj;
        String typeLabel = Channel.getChannel().getTypeLabel(type, getWorkspace(), getUserLang());
        String label = JcmsUtil.getHtmlIcon(JcmsUtil.getClassIcon(type)) + " " + typeLabel;
        sb.append("<li>").append(generateAddLink(type, label, disabled)).append("</li>");
      }
      sb.append("</ul>");

      sb.append("</li>");
    } else { // just one
      Class<? extends Data> clazz = (Class<? extends Data>) typeSet.iterator().next();
      String typeLabel = Channel.getChannel().getTypeLabel(clazz, getWorkspace(), getUserLang());
      String label = JcmsUtil.getHtmlIcon("add") + " " + JcmsUtil.glp(getUserLang(), "ui.com.alt.add") + " " + typeLabel;
      sb.append("<li class=\"add-item").append(hidden ? " hide" : "").append("\">");
      sb.append(generateAddLink(clazz, label, disabled));
      sb.append("</li>");
    }

    return sb.toString();
  }

  private String generateAddLink(Class<? extends Data> clazz, String label, boolean disabled) {

    TypeTemplateEntry tte = Publication.getTemplateEntry(TypeTemplateEntry.USAGE_EDIT_POPUP, false, clazz, getWorkspace());
    String editTemplatePath = tte != null ? tte.getPath() : "edit.jsp";

    Map<String, String[]> paramsMap = new HashMap<String, String[]>();
    paramsMap.put("popupEdition", new String[] { "true" });
    paramsMap.put("redirect", new String[] { "$CHOOSER$" });
    String addUrl = URLUtils.buildUrl(editTemplatePath, paramsMap);

    StringBuilder sb = new StringBuilder();
    sb.append("<a href=\"#\"");
    if (FileDocument.class.isAssignableFrom(clazz)) {
      sb.append(" data-jalios-action=\"chooser:document\"");
      String options = "type=" + clazz.getCanonicalName() + "&" + getOption(ChooserConstants.CHOOSER_OPTION, "");
      sb.append(" data-jalios-options=\"" + options + "\"");
    } else {
      sb.append(" data-jalios-action=\"chooser:add\" data-jalios-chooser=\"").append(addUrl).append("\"");
    }

    sb.append(disabled ? " disabled" : "");
    sb.append(" onclick=\"return false;\">");
    sb.append(label);
    sb.append("</a>");

    return sb.toString();
  }

}
