package com.jalios.jcms.uicomponent;

import java.util.HashMap;
import java.util.Map;

import com.jalios.jcms.HttpUtil;
import com.jalios.util.Util;

public class DataAttribute {

  public static final String DATA_ATTR_PREFIX = "data-";

  Map<String, Object> dataAttributesMap = new HashMap<String, Object>();

  public DataAttribute addData(String dataName, Object dataValue) {
    return setData(Util.getHashMap(dataName, dataValue));
  }

  public DataAttribute setData(Map<String, Object> dataAttributes) {
    if (dataAttributes != null) {
      for (Map.Entry<String, Object> entry : dataAttributes.entrySet()) {
        String dataName = entry.getKey();
        Object dataValue = entry.getValue();
        dataAttributesMap.put(dataName.startsWith(DATA_ATTR_PREFIX) ? dataName.substring(DATA_ATTR_PREFIX.length()) : dataName, dataValue);
      }
    }
    return this;
  }

  public Object getData(String dataName) {
    String name = dataName.startsWith(DATA_ATTR_PREFIX) ? dataName.substring(DATA_ATTR_PREFIX.length()) : dataName;
    return dataAttributesMap.get(name);
  }

  public boolean containsData(String dataName) {
    String name = dataName.startsWith(DATA_ATTR_PREFIX) ? dataName.substring(DATA_ATTR_PREFIX.length()) : dataName;
    return dataAttributesMap.containsKey(name);
  }

  @Override
  public String toString() {
    return formatDataAttributes();
  }

  private String formatDataAttributes() {
    StringBuilder sb = new StringBuilder();
    if (dataAttributesMap != null) {
      for (Map.Entry<String, Object> entry : dataAttributesMap.entrySet()) {
        String dataName = entry.getKey();
        String dataValue = HttpUtil.encodeForHTMLAttribute(Util.getString(entry.getValue(), ""));
        if (sb.length() > 0) {
          sb.append(" ");
        }
        sb.append(!dataName.startsWith(DATA_ATTR_PREFIX) ? DATA_ATTR_PREFIX : "").append(dataName).append("='").append(dataValue).append("'");
      }
    }
    return sb.toString();
  }
}
